#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
whistler_b2e.py

@author : Louis RICHARD
"""

import numpy as np
import xarray as xr

from astropy import constants

from ..pyrf.plasma_calc import plasma_calc


def whistler_b2e(b2=None, freq=None, theta_k=None, b_mag=None, n_e=None):
    """Computes electric field power as a function of frequency for whistler waves using magnetic field power and cold
    plasma theory.

    Parameters
    ----------
    b2 : xarray.DataArray
        Time series of the power of whistler magnetic field in nT^2 Hz^{-1}.

    freq : numpy.ndarray
        frequencies in Hz corresponding B2.

    theta_k : float
        wave-normal angle of whistler waves in radians.

    b_mag : xarray.DataArray
        Time series of the magnitude of the magnetic field in nT.

    n_e : xarray.DataArray
        Time series of the electron number density in cm^{-3}.

    Returns
    -------
    e2 : xarray.DataArray
        Time series of the electric field power.

    Examples
    --------
    >>> from pyrfu import mms
    >>> e_power = mms.whistler_b2e(b_power, freq, theta_k, b_mag, n_e)

    """

    assert b2 is not None and isinstance(b2, xr.DataArray)
    assert freq is not None and isinstance(freq, np.ndarray)
    assert theta_k is not None and isinstance(theta_k, float)
    assert b_mag is not None and isinstance(b_mag, xr.DataArray)
    assert n_e is not None and isinstance(n_e, xr.DataArray)

    # Calculate plasma parameters
    pparam = plasma_calc(b_mag, n_e, n_e, n_e, n_e)
    fpe, fce = [pparam.Fpe, pparam.Fce]

    c = constants.c.value

    # Check input
    if len(b2) != len(freq):
        raise IndexError("B2 and freq lengths do not agree!")

    # Calculate cold plasma parameters
    r = 1 - fpe ** 2 / (freq * (freq - fce))
    l = 1 - fpe ** 2 / (freq * (freq + fce))
    p = 1 - fpe ** 2 / freq ** 2
    d = 0.5 * (r - l)
    s = 0.5 * (r + l)

    n2 = r * l * np.sin(theta_k) ** 2
    n2 += p * s * (1 + np.cos(theta_k) ** 2)
    n2 -= np.sqrt((r * l - p * s) ** 2 * np.sin(theta_k) ** 4 + 4 * (p ** 2) * (d ** 2) * np.cos(
        theta_k) ** 2)
    n2 /= (2 * (s * np.sin(theta_k) ** 2 + p * np.cos(theta_k) ** 2))

    e_temp1 = (p - n2 * np.sin(theta_k) ** 2) ** 2. * ((d / (s - n2)) ** 2 + 1) + (
            n2 * np.cos(theta_k) * np.sin(theta_k)) ** 2

    e_temp2 = (d / (s - n2)) ** 2 * (p - n2 * np.sin(theta_k) ** 2) ** 2 + p ** 2 * np.cos(
        theta_k) ** 2

    e2 = (c ** 2 / n2) * e_temp1 / e_temp2 * b2 * 1e-12

    return e2



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
.py

@author : Louis RICHARD
"""

import numpy as np
import xarray as xr

from ..pyrf import calc_dt


def psd_rebin(vdf=None, phi=None, energy0=None, energy1=None, step_table=None):
    """Convert burst mode distribution into 64 energy channel distribution.
    Functions takes the burst mode distribution sampled in two energy tables and converts to a
    single energy table with 64 energy channels. Time resolution is halved and phi angles are
    averaged over adjacent times.

    Parameters
    ----------
    vdf : xarray.Dataset
        Time series of the particle distribution.

    phi : xarray.DataArray
        Time series of the phi angles.

    energy0 : xarray.DataArray or numpy.ndarray
        Energy table 0.

    energy1 : xarray.DataArray or numpy.ndarray
        Energy table 1.

    step_table : xarray.DataArray
        Time series of the stepping table between energies (burst).

    Returns
    -------
    vdf_r : numpy.ndarray
        Rebinned particle distribution.

    phi_r : numpy.ndarray
        Time series of the recalculated phi angle.

    energy_r : numpy.ndarray
        Revised energy table.

    Notes
    -----
    I'm assuming no gaps in the burst data interval. If there is a gap use time_clip before running.
    To be updated later.
    
    """

    assert vdf is not None
    assert phi is not None
    assert energy0 is not None and isinstance(energy0, (xr.DataArray, np.ndarray))
    assert energy1 is not None and isinstance(energy1, (xr.DataArray, np.ndarray))
    assert step_table is not None

    if isinstance(energy0, xr.DataArray):
        energy0 = energy0.data
    else:
        pass

    if isinstance(energy1, xr.DataArray):
        energy1 = energy1.data
    else:
        pass

    step_table = step_table.data

    # Sort energy levels
    energy_r = np.sort(np.hstack([energy0, energy1]))

    # Define new times
    dt = calc_dt(vdf.data)
    new_time = vdf.time.data[:-1:2] + int(dt * 1e9 / 2)

    vdf_r, phi_r = [np.zeros((len(new_time), 64, 32, 16)), np.zeros((len(new_time), 32))]

    phi_s = np.roll(phi.data, 2, axis=1)
    phi_s[:, 0] = phi_s[:, 0] - 360

    time_indices = np.arange(0, len(vdf.time), 2)

    for new_el_num, ii in enumerate(time_indices):
        if phi.data[ii, 0] > phi.data[ii + 1, 0]:
            phi_r[new_el_num, :] = (phi.data[ii, :] + phi_s[ii + 1, :]) / 2

            vdf_temp = np.roll(np.squeeze(vdf.data.data[ii + 1, ...]), 2, axis=1)

            if step_table[ii]:
                vdf_r[new_el_num, 1:64:2, ...] = vdf.data.data[ii, ...]
                vdf_r[new_el_num, 0:63:2, ...] = vdf_temp
            else:
                vdf_r[new_el_num, 1:64:2, ...] = vdf.data.data[ii, ...]
                vdf_r[new_el_num, 0:63:2, ...] = vdf_temp

        else:
            phi_r[new_el_num, :] = (phi.data[ii, :] + phi.data[ii + 1, :]) / 2

            if step_table[ii]:
                vdf_r[new_el_num, 1:64:2, ...] = vdf.data.data[ii, ...]
                vdf_r[new_el_num, 0:63:2, ...] = vdf.data.data[ii + 1, ...]
            else:
                vdf_r[new_el_num, 1:64:2, ...] = vdf.data.data[ii + 1, ...]
                vdf_r[new_el_num, 0:63:2, ...] = vdf.data.data[ii, ...]

    return vdf_r, phi_r, energy_r



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
lh_wave_analysis.py

@author : Louis RICHARD
"""

import numpy as np
import xarray as xr

from astropy import constants

from ..pyrf.filt import filt
from ..pyrf.calc_dt import calc_dt
from ..pyrf.resample import resample
from ..pyrf.convert_fac import convert_fac
from ..pyrf.ts_scalar import ts_scalar
from ..pyrf.extend_tint import extend_tint
from ..pyrf.time_clip import time_clip
from ..pyrf.ts_vec_xyz import ts_vec_xyz


def lh_wave_analysis(tints=None, e_xyz=None, b_scm=None, b_xyz=None, n_e=None, **kwargs):
    """
    Calculates lower-hybrid wave properties from MMS data

    Parameters
    ----------
    tints : list of str
        Time interval

    e_xyz : xarray.DataArray
        Time series pf the electric field

    b_scm : xarray.DataArray
        Time series of the fluctuations of the magnetic field

    b_xyz : xarray.DataArray
        Time series of the background magnetic field

    n_e : xarray.DataArray
        Time series of the number density

    **kwargs : dict
        Hash table of keyword arguments with :
            * lhfilt : float or int or list of float or list of int
                Filter for LH fluctuations. For one element it is the minimum frequency in the
                highpass filter. For two elements the fields are bandpassed between the frequencies.

            * blpass : float or int
                Set maximum frequency for low-pass filter of background magnetic field (FGM)

    Returns
    -------
    phi_eb : xarray.DataArray
        to fill

    v_best : numpy.ndarray
        to fill

    dir_best : numpy.ndarray
        to fill

    thetas : numpy.ndarray
        to fill

    corrs : numpy.ndarray
        to fill

    Examples
    --------
    >>> from pyrfu.mms import get_data, lh_wave_analysis
    >>> # Large time interval
    >>> tint_long = ["2015-12-14T01:17:39.000", "2015-12-14T01:17:43.000"]
    >>> # Load fields and density
    >>> b_xyz = get_data("B_gse_fgm_brst_l2", tint_long, 2)
    >>> e_xyz = get_data("E_gse_edp_brst_l2", tint_long, 2)
    >>> b_scm = get_data("B_gse_scm_brst_l2", tint_long, 2)
    >>> n_e = get_data("Ne_fpi_brst_l2", tint_long, 2)
    >>> # Time interval of focus
    >>> tint = ["2015-12-14T01:17:40.200","2015-12-14T01:17:41.500"]
    >>> opt = dict(lhfilt=[5, 100])
    >>> phi_eb, v_best, dir_best, theta, _ = lh_wave_analysis(tint, e_xyz, b_scm, b_xyz, n_e, **opt)

    """

    # Default band passes
    min_freq = 10
    max_freq = 0
    lowpass_b_xyz = 2

    if "lhfilt" in kwargs:
        if isinstance(kwargs["lhfilt"], (float, int)):
            min_freq = kwargs["lhfilt"]
        elif isinstance(kwargs["lhfilt"], (list, np.ndarray)) and len(kwargs["lhfilt"]):
            min_freq = kwargs["lhfilt"][0]
            max_freq = kwargs["lhfilt"][1]
        else:
            raise ValueError("lhfilt option not recognized")

    if "blpass" in kwargs:
        if isinstance(kwargs["blpass"], (float, int)):
            lowpass_b_xyz = kwargs["blpass"]
        else:
            raise ValueError("blpass option not recognized")

    # Bandpass filter data
    b_xyz = filt(b_xyz, 0, lowpass_b_xyz, 5)
    e_xyz = resample(e_xyz, b_scm)
    n_e = resample(n_e, b_scm)
    b_xyz = resample(b_xyz, b_scm)
    b_scm_fac = convert_fac(b_scm, b_xyz, [1, 0, 0])
    b_scm_fac = filt(b_scm_fac, min_freq, max_freq, 5)
    e_xyz = filt(e_xyz, min_freq, max_freq, 5)

    qe, mu0 = [constants.e.value, constants.mu0.value]

    b_mag = np.linalg.norm(b_xyz, axis=1)
    phi_b = (b_scm_fac.data[:, 2]) * b_mag * 1e-18 / (n_e.data * qe * mu0 * 1e6)
    phi_b = ts_scalar(b_scm_fac.time.data, phi_b)

    # short buffer so phi_E does not begin at zero.
    tint = extend_tint(tints, [-.2, .2])

    e_xyz = time_clip(e_xyz, tint)
    phi_bs = time_clip(phi_b, tints)

    # Rotate e_xyz into field-aligned coordinates
    b_xyz_tc = time_clip(b_xyz, tints)
    b_mean = np.mean(b_xyz_tc.data, axis=0)
    b_vec = b_mean / np.linalg.norm(b_mean)
    r_temp = [1, 0, 0]
    r2 = np.cross(b_vec, r_temp)
    r2 /= np.linalg.norm(r2)
    r1 = np.cross(r2, b_vec)
    er1 = e_xyz.data[:, 0] * r1[0] + e_xyz.data[:, 1] * r1[1] + e_xyz.data[:, 2] * r1[2]
    er2 = e_xyz.data[:, 0] * r2[0] + e_xyz.data[:, 1] * r2[1] + e_xyz.data[:, 2] * r2[2]
    er3 = e_xyz.data[:, 0] * b_vec[0] + e_xyz.data[:, 1] * b_vec[1] + e_xyz.data[:, 2] * b_vec[2]

    e_fac = ts_vec_xyz(e_xyz.time.data, np.vstack([er1, er2, er3]).T)

    # Find best direction
    dt_e_fac = calc_dt(e_fac)
    thetas = np.linspace(0, 2 * np.pi, 361)
    corrs = np.zeros(len(thetas))

    for ii, theta in enumerate(thetas):
        e_temp = np.cos(theta) * e_fac.data[:, 0] + np.sin(theta) * e_fac.data[:, 1]

        phi_temp = ts_scalar(e_xyz.time.data, np.cumsum(e_temp) * dt_e_fac)
        phi_temp = time_clip(phi_temp, tints)
        phi_temp -= np.mean(phi_temp)

        corrs[ii] = np.corrcoef(phi_bs.data, phi_temp.data)

    corrpos = np.argmax(corrs)
    e_best = np.cos(thetas[corrpos]) * e_fac.data[:, 0] + np.sin(thetas[corrpos]) * e_fac.data[:, 1]
    e_best = ts_scalar(e_xyz.time.data, e_best)
    phi_best = ts_scalar(e_xyz.time.data, np.cumsum(e_best) * dt_e_fac)
    phi_best = time_clip(phi_best, tints)
    phi_best -= np.mean(phi_best)
    theta_best = thetas[corrpos]
    dir_best = r1 * np.cos(theta_best) + r2 * np.sin(theta_best)

    # Find best speed
    # Maximum velocity may need to be increased in rare cases
    vph_vec = np.linspace(1e1, 5e2, 491)
    corr_v = np.zeros(len(vph_vec))

    for ii, vph in enumerate(vph_vec):
        phi_e_temp = phi_best.data * vph
        corr_v[ii] = np.sum(np.abs(phi_e_temp - phi_bs.data) ** 2)

    corr_vpos = np.argmin(corr_v)
    phi_e_best = phi_best.data * vph_vec[corr_vpos]
    phi_e_best = ts_scalar(phi_bs.time.data, phi_e_best)
    v_best = vph_vec[corr_vpos]

    phi_eb = xr.DataArray(np.vstack([phi_e_best.data, phi_bs.data]).T,
                          coords=[phi_bs.time, ["Ebest", "Bs"]], dims=["time", "comp"])

    return phi_eb, v_best, dir_best, thetas, corrs



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
ts_append.py

@author : Louis RICHARD
"""

import numpy as np
import xarray as xr


def ts_append(inp1=None, inp2=None):
    """Concatenate two time series along the time axis.

    Parameters
    ----------
    inp1 : xarray.DataArray
        Time series of the first input (early times).

    inp2 : xarray.DataArray
        Time series of the second input (late times).

    Returns
    -------
    out : xarray.DataArray
        Concatenated time series.

    Notes
    -----
    The time series must be in the correct time order.

    """

    assert inp1 is None or isinstance(inp1, xr.DataArray)
    assert inp2 is not None and isinstance(inp2, xr.DataArray)

    if inp1 is None:
        return inp2

    out_data = {}

    if inp1.data.ndim != 1:
        out_data["data"] = np.vstack([inp1, inp2])

    else:
        out_data["data"] = np.hstack([inp1, inp2])

    out_data["attrs"] = {}

    for k in inp1.attrs:
        if isinstance(inp1.attrs[k], np.ndarray):
            out_data["attrs"][k] = np.hstack([inp1.attrs[k], inp2.attrs[k]])

        else:
            out_data["attrs"][k] = inp1.attrs[k]

    depends = [{} for _ in range(len(inp1.dims))]

    for i, dim in enumerate(inp1.dims):
        if i == 0 or dim == "time":
            depends[i]["data"] = np.hstack([inp1[dim].data, inp2[dim].data])

            # add attributes
            depends[i]["attrs"] = {}

            for k in inp1[dim].attrs:
                # if attrs is array time append
                if isinstance(inp1[dim].attrs[k], np.ndarray):
                    depends[i]["attrs"][k] = np.hstack([inp1[dim].attrs[k], inp2[dim].attrs[k]])

                else:
                    depends[i]["attrs"][k] = inp1[dim].attrs[k]

        else:
            # Use values of other coordinates of inp1 assuming equal to inp2
            depends[i]["data"] = inp1[dim].data

            # add attributes
            depends[i]["attrs"] = {}

            for k in inp1[dim].attrs:
                depends[i]["attrs"][k] = inp1[dim].attrs[k]

    # Prepare coord and dims to build DataArray
    dims, coord = [inp1.dims, [None] * len(inp1.dims)]

    for i, depend in enumerate(depends):
        coord[i] = depend["data"]

    # Create DataArray
    out = xr.DataArray(out_data["data"], coords=coord, dims=dims, attrs=out_data["attrs"])

    for i, dim in enumerate(dims):
        out[dim].attrs = depends[i]["attrs"]

    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
fk_power_spectrum.py

@author : Louis RICHARD
"""

import numpy as np
import xarray as xr


def fk_power_spectrum(pc=-1, tint=None, v6=None, b_xyz=None, z_phase=None, cav=128, field=True,
                      num_k=101, num_f=200, df=None):
    """Calculates the frequency-wave number power spectrum using MMS's four spin plane probes.
    Follows the same procedure as c_wk_powerspec. Function uses L2 scpot probe potential data to
    construct electric fields aligned with B separated by 60 m. Wavelet based cross-spectral
    analysis is used to calculate the phase difference between and the fields, and hence the wave
    number. The power is then binned according to frequency and wave number
    (Graham et al., JGR, 2016).

    Parameters
    ----------
    pc : int
        Probe combination to use :
        * 1 : ``b_xyz`` aligned with probes 1 and 2.
        * 3 : ``b_xyz`` aligned with probes 3 and 4.
        * 5 : ``b_xyz`` aligned with probes 5 and 6.

    tint : list of str
        Time interval over which the power spectrum is calculated. ``b_xyz`` should be closely
        aligned with one probe pair over this time.

    v6 : xarray.DataArray
        L2 probe potentials.

    b_xyz : xarray.DataArray
        Time series of the magnetic field in DMPA coordinates.

    z_phase : xarray.DataArray
        Spacecraft phase (zphase). Obtained from ancillary_defatt. Needed only if ``pc`` = 1 or 3.

    cav : int, optional
        Number of points in time series used to estimate phase. Default is ``cav`` = 128.

    field : bool, optional
        Set to True to use electric fields calculated from opposing probes and sc_pot. Set to
        False to use only opposing probe potentials. Default is True.

    num_k : int, optional
		Set number of wave numbers used in spectrogram. Default ``num_k`` = 101.

	df : float, optional
		Linearly spaced frequencies. Default is logarithmic spacing.

	num_f : int, optional
		Set number of frequencies used in spectrogram. Default ``num_f`` = 200.

	Returns
	-------
	fk_power : xarray.DataArray
	    Powers as a function of frequency and wavenumber. Power is normalized to the maximum value.

    See also
    --------
    pyrfu.mms.probe_align_times : Returns times when f-a electrostatic waves can be characterized.

    Notes
    -----
    Timing corrections are applied to ``v6`` in this function. Do not apply them before running
    this function. Directions and speeds are consistent with expectations based on time delays.
    Work still in progress. Time corrections for the probes need to be revised.

    Examples
    --------
    >>> from pyrfu import mms
    >>> fk_power = mms.fk_power_spectrum(pc, tint, v6, b_xyz, z_phase, cav=256, field=False)
    >>> fk_power = mms.fk_power_spectrum(pc, tint, v6, b_xyz, z_phase, cav=256, df=50, num_k=500)

    """

    # Check input
    assert isinstance(pc, int) and pc in [1, 3, 5]
    assert isinstance(tint, list) and isinstance(tint[0], str) and isinstance(tint[1], str)
    assert v6 is not None and isinstance(v6, xr.DataArray)
    assert b_xyz is not None and isinstance(b_xyz, xr.DataArray)
    assert z_phase is not None and isinstance(z_phase, xr.DataArray)
    assert isinstance(cav, int)
    assert isinstance(field, bool)
    assert isinstance(num_k, int)
    assert isinstance(num_f, int)

    return



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
medfilt.py

@author : Louis RICHARD
"""

import xarray as xr
import numpy as np
from scipy import signal


def medfilt(inp=None, n_pts=11):
    """Applies a median filter over npts points to inp.

    Parameters
    ----------
    inp : xarray.DataArray
        Time series of the input variable.

    n_pts : float or int
        Number of points of median filter.

    Returns
    -------
    out : xarray.DataArry
        Time series of the median filtered input variable.

    Examples
    --------
    >>> import numpy
    >>> from pyrfu import mms, pyrf
    >>> # Time interval
    >>> tint = ["2019-09-14T07:54:00.000", "2019-09-14T08:11:00.000"]
    >>> # Spacecraft indices
    >>> mms_list = numpy.arange(1,5)
    >>> # Load magnetic field and electric field
    >>> r_mms, b_mms = [[] * 4 for _ in range(2)]
    >>> for mms_id in range(1, 5):
    >>> 	r_mms.append(mms.get_data("R_gse", tint, mms_id))
    >>> 	b_mms.append(mms.get_data("B_gse_fgm_srvy_l2", tint, mms_id))
    >>>
    >>> # Compute current density, etc
    >>> j_xyz, div_b, b_xyz, jxb, div_t_shear, div_pb = pyrf.c_4_j(r_mms, b_mms)
    >>> # Get J sampling frequency
    >>> fs = pyrf.calc_fs(j_xyz)
    >>> # Median filter over 1s
    >>> j_xyz = pyrf.medfilt(j_xyz,fs)

    """

    assert inp is not None and isinstance(inp, xr.DataArray)
    assert isinstance(n_pts, (float, int))

    if isinstance(n_pts, float):
        n_pts = np.floor(n_pts).astype(int)

    if n_pts % 2 == 0:
        n_pts += 1

    nt = len(inp)

    if inp.ndim == 3:
        inp_data = np.reshape(inp.data, [nt, 9])
    else:
        inp_data = inp.data

    try:
        n_comp = inp_data.shape[1]
    except IndexError:
        n_comp = 1

        inp_data = inp_data[..., None]

    out_data = np.zeros(inp_data.shape)

    if not n_pts % 2:
        n_pts += 1

    for i in range(n_comp):
        out_data[:, i] = signal.medfilt(inp_data[:, i], n_pts)

    if n_comp == 9:
        out_data = np.reshape(out_data, [nt, 3, 3])

    if out_data.shape[1] == 1:
        out_data = out_data[:, 0]

    out = xr.DataArray(out_data, coords=inp.coords, dims=inp.dims)
    out.attrs = inp.attrs

    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
.py

@author : Louis RICHARD
"""

def generic_minimum_residue_analysis_engine(eta=None, q=None, constraint_vector=None):
    """
    Calculates L, V and U given the density of conserved quantity eta and transport tensor q
    using constraint that discontinuity normal is perpendicular to the constraint vector.

    Implements general Generic Minimum Residue Analysis (GMRA) solution based on [8]_, [9]_

    Parameters
    ----------
    eta : xarray.DataArray
        Density of conserved quantity Mx3, M is number of points

    q : xarray.DataArray
        Transport tensor

    constraint_vector : list or numpy.ndarray
        Constraint vector

    Returns
    -------
    l :
        Eigenvalues

    v :
        Eigenvectors corresponding to lmin, lmean, lmax

    u :
        Transport velocity, boundary velocity is Un = U * V(:, 1)

    References
    ----------
    .. [8]  Sonnerup, B. U. Ö., S. Haaland, G. Paschmann, M. W. Dunlop, H. Rème, and A. Balogh (
            2006), Orientation and motion ofa plasma discontinuity from single-spacecraft
            measurements : Generic residue analysis of Cluster data, J. Geophys. Res., 111,
            A05203,doi : https://doi.org/10.1029/2005JA011538 .

    .. [9]  Sonnerup, B. U. Ö., S. Haaland, G. Paschmann, M. W. Dunlop, H. Rème, and A. Balogh (
            2007), Correction to ‘Orientationand motion of a plasma discontinuity from
            single-spacecraft measurements : Generic residue analysis of Cluster data´,
            J. Geophys.Res., 11 2 , A04201, doi : hhtps://doi.org/10.1029/2007JA012288 .

    """

    # Defaults
    do_calculate_velocity, do_calculate_q, have_constraint = [True, True, False]

    # Input check

    if constraint_vector is not None:
        have_constraint = True

    # Check inputs
    if eta.ndim == 1:
        eta = np.tile(eta, (len(q), 1))

    # U estimate. U = < deta dq > / < | deta | ^ 2 >
    d_eta = eta - np.namean(eta, axis=0)  # Eq. 10
    d_q = q - np.namean(q, axis=0)
    #d_eta_d_q_average = np.nanmean(matrix_dot(deta, 1, dq, 1), 1)
    #d_eta2_average = np.nanmean(np.dot(deta, deta, 2), 1)
    u = d_eta_d_q_average / d_eta2_averag  # Eq. 12

    # Q estimate
    # dq_dq_average = shiftdim(np.nanmean(matrix_dot(dq, 1, dq, 1), 1), 1)
    # d_eta_d_q_average2_mat = detadqAver.T * detadqAver
    if eta == 0:
        # Eq.19
        q_mat = dq_dq_average
    else:
        # Eq.15b (see correction in Sonnerup 2007)
        q_mat = dq_dq_average - d_eta_dq_aver2_mat / d_eta2_average

    # Correct Q by the number of dimensions so that eigenvalues coorespond to the variance of dq
    q_mat = q_mat / d_q.shape[1]


    """
    if q is not None:
        
    
    if nargin == 0 & & nargout == 0
        help irf_generic_minimum_residue_analysis_engine
        return;
    elseif nargin == 1 && isstruct(varargin{1})
        InputParameters = varargin{1};
        inputParameterFields = fieldnames(InputParameters);
        for j in range(len(input_parameter_field)):
            fieldname = input_parameter_fields{j};
            eval([fieldname ' = InputParameters.' fieldname ';']);
        end
    elseif nargin > 1
    
    
        args = varargin;
        while numel(args) >= 2
            switch args{1}
                case 'Q'
                    Q = args{2};
                    doCalculateVelocity = false;
                    doCalculateQ = false;
                case 'eta'
                    eta = args{2};
                case 'q'
                    q = args{2};
                case 'constraint'
                    constraintVector = args{2};
                    haveConstraint = true;
                otherwise
                    irf.log('critical', 'unrecognized input');
                    return
    
            args(1: 2)=[];
    
    
    # Check inputs
    if numel(eta) == 1 % scalar
        eta = repmat(eta, size(q, 1), 1);
    end
    
    # Calculate Q from eta and q
        if doCalculateQ
            # U estimate. U = < deta dq > / < | deta | ^ 2 > deta = bsxfun( @ minus, eta, irf.nanmean(eta, 1)); % Eq. 10
            dq = bsxfun( @ minus, q, irf.nanmean(q, 1))
            deta_dq_aver = np.nanmean(matrix_dot(deta, 1, dq, 1), 1)
            deta2_aver = np.nanmean(np.dot(deta, deta, 2), 1)
            U = detadqAver / deta2Aver  # Eq. 12
    
            # Q estimate
            dqdqAver = shiftdim(np.nanmean(matrix_dot(dq, 1, dq, 1), 1), 1)
            detadqAver2Matrix = detadqAver.T * detadqAver
            if eta == 0:
                Q = dq_dq_aver # Eq.19
            else:
                Q = dq_dq_aver - deta_dq_aver2_matrix / deta2_aver # Eq.15b (see correction in Sonnerup 2007)
    
    
            # Correct Q by the number of dimensions so that eigenvalues coorespond to the variance of dq
            Q = Q / size(dq, 2)
    
        # Check for constraints
            if haveConstraint
                P = eye(numel(constraintVector)) - constraintVector(:) * constraintVector(:)'; % Eq 41
                Q = P * Q * P
    
    
        # Calculate eigenvalues and eigenvectors from Q
        [V, lArray] = eig(Q)            # L is diagonal matrix of eigenvalues and V matrix with columns eigenvectors
        [L, I] = sort(diag(lArray))
        V = V(:, I)
    
        if ~doCalculateVelocity
            U = NaN
            return
        end
    
        # Calculate normal velocity
    
        Un = dot(U, V(:, 1))
    
        # Print output
        if nargout == 0
            disp(['Eigenvalues: ' sprintf('%7.5f ', L)]);
            disp(vector_disp('N', V(:, 1)));
            disp(vector_disp('M', V(:, 2)));
            disp(vector_disp('L', V(:, 3)));
            disp(vector_disp('U', U, 'km/s'));
            disp(['Un = ' num2str(Un, 3), ' km/s']);
        end
    
        % % Define
        output
        if nargout == 0
            clear
            L
            V
            U;
        end
    
        % % Functions
        function out = vector_disp(vectSymbol, vect, vectUnit)
        if nargin == 2, vectUnit ='';end
        out = sprintf(['|' vectSymbol '| = %8.4f ' vectUnit...
                       ', ' vectSymbol ' = [ %8.4f %8.4f %8.4f ] ' vectUnit], ...
        norm(vect), vect(1), vect(2), vect(3));
    
    function out = matrix_dot(inp1, ind1, inp2, ind2)
    % MATRIX_DOT summation over one index multiplication
    %
    % MATRIX_DOT(inp1, ind1, inp2, ind2)
    % inp1, inp2 are the matrixes and summation is over dimensions(ind1 + 1) and
    % (ind2 + 1). + 1 because first dimension is always time.
    szinp1 = size(inp1);
    ndimsInp1 = ndims(inp1) - 1;
    szinp2 = size(inp2);
    ndimsInp2 = ndims(inp2) - 1;
    szout1 = szinp1;
    szout1(ind1 + 1) = [];
    szout2 = szinp2;
    szout2([1 ind2 + 1]) = [];
    szout = [szout1 szout2];
    out = zeros(szout);
    
    if ndimsInp1 == 1
        if ndimsInp2 == 1
            out = sum(inp1. * inp2, 2);
        elseif
            ndimsInp2 == 2 & & ind2 == 1
            for jj = 1:szinp2(3)
                for kk = 1:szinp1(2)
                    out(:, jj) = out(:, jj) + inp1(:, kk).*inp2(:, kk, jj);
                end
            end
        elseif
            ndimsInp2 == 2 & & ind2 == 2
            for jj = 1:szinp2(2)
                for kk = 1:szinp1(2)
                    out(:, jj) = out(:, jj) + inp1(:, kk).*inp2(:, jj, kk);
                end
            end
        else
            error('Not yet implemented'); % not implemented
        end
    elseif
        ndimsInp1 == 2 & & ndimsInp2 == 1 & & ind1 == 1
        for jj = 1:szinp1(2)
            for kk = 1:szinp2(2)
                out(:, jj) = out(:, jj) + inp1(:, kk, jj).*inp2(:, kk);
            end
        end
    elseif
        ndimsInp1 == 2 & & ndimsInp2 == 1 & & ind1 == 2
        for jj = 1:szinp1(2)
            for kk = 1:szinp2(2)
                out(:, jj) = out(:, jj) + inp1(:, jj, kk).*inp2(:, kk);
            end
        end
    elseif
        ndimsInp1 == 2 & & ndimsInp2 == 2 & & ind1 == 1 & & ind2 == 1
        for jj = 1:szinp1(3)
            for kk = 1:szinp2(3)
                for ss = 1:szinp1(ind1 + 1)
                    out(:, jj, kk) = out(:, jj, kk) + inp1(:, ss, jj).*inp2(:, ss, kk);
                end
            end
        end
    else
        error('Not yet implemented'); % not implemented
    end
    """


#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
mean.py

@author : Louis RICHARD
"""

import numpy as np
import xarray as xr

from . import resample, cross, norm, ts_vec_xyz


def mean(inp=None, r=None, b=None, z=None):
    """Put inp into mean field coordinates defined by position vector r and magnetic field b
    if earth magnetic dipole axis z is given then uses another algorithm (good for auroral passages)

    Parameters
    ----------
    inp : xarray.DataArray
        Input field to put into MF coordinates.

    r : xarray.DataArray
        Time series of the spacecraft position.

    b : xarray.DataArray
        Time series of the background magnetic field.

    z : xarray.DataArray
        Earth magnetic dipole axis.

    Returns
    -------
    out : xarray.DataArray
        Input field in mean field coordinates.

    """

    assert inp is not None and isinstance(inp, xr.DataArray)
    assert r is not None and isinstance(r, xr.DataArray)
    assert b is not None and isinstance(b, xr.DataArray)

    #
    if z is not None:
        flag_dipole = True

        if isinstance(z, xr.DataArray):
            if len(z) != len(inp):
                zz = resample(z, inp)
            else:
                zz = z
        else:
            raise TypeError("z must be a DataArray")

    else:
        flag_dipole = False
        zz = None

    if len(r) != len(inp):
        rr = resample(r, inp)
    else:
        rr = r

    if len(b) != len(inp):
        bb = resample(b, inp)
    else:
        bb = b

    zv = norm(bb)

    if not flag_dipole:
        yv = cross(zv, rr)
        yv /= np.linalg.norm(yv, axis=1)[:, None]
    else:
        ss = np.sum(b * r)
        ind = ss > 0
        ss = -1 * np.ones(ss.shape)
        ss[ind] = 1
        yv = np.cross(zz, bb) * ss[:, None]
        yv /= np.linalg.norm(yv, axis=1)[:, None]

    xv = np.cross(yv, zv)

    # in case rotation axis is used as reference uncomment next line
    # rot_axis=rr
    # rot_axis(:,[2 3])=0
    # yv=irf_norm(irf_cross(irf_cross(bb,rot_axis), bb));
    # xv=irf_cross(yv,zv);

    out_data = np.zeros(inp.data.shape)
    out_data[:, 0] = np.sum(xv * inp, axis=1)
    out_data[:, 1] = np.sum(yv * inp, axis=1)
    out_data[:, 2] = np.sum(zv * inp, axis=1)

    out = ts_vec_xyz(inp.time.data, out_data, inp.attrs)

    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
ts_tensor_xyz.py

@author : Louis RICHARD
"""

import numpy as np
import xarray as xr


def ts_tensor_xyz(t=None, data=None, attrs=None):
    """Create a time series containing a 2nd order tensor.

    Parameters
    ----------
    t : numpy.ndarray
        Array of times.

    data : numpy.ndarray
        Data corresponding to the time list.

    attrs : dict
        Attributes of the data list.

    Returns
    -------
    out : xarray.DataArray
        2nd order tensor time series.

    """

    assert t is not None and isinstance(t, np.ndarray)
    assert data is not None and isinstance(data, np.ndarray)
    assert data.ndim == 3 and data.shape[1:] == (3, 3)

    # Check inputs are not empty
    if t is None:
        raise ValueError("ts_tensor_xyz requires at least two inputs")

    if data is None:
        raise ValueError("ts_tensor_xyz requires at least two inputs")

    # Check inputs are numpy arrays
    if len(t) != len(data):
        raise IndexError("Time and data must have the same length")

    flag_attrs = True

    if attrs is None:
        flag_attrs = False

    out = xr.DataArray(data, coords=[t[:], ["x", "y", "z"], ["x", "y", "z"]],
                       dims=["time", "comp_h", "comp_v"])

    if flag_attrs:
        out.attrs = attrs

        out.attrs["TENSOR_ORDER"] = 2
    else:
        out.attrs["TENSOR_ORDER"] = 2

    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
wave_fft.py

@author : Louis RICHARD
"""

import numpy as np
import xarray as xr

from scipy import signal


def wave_fft(x=None, window="hamming", frame_overlap=10, frame_length=20, fs=None):
    """Short-Time Fourier Transform.

    Parameters
    ----------
    x : xarray.DataArray
        Time series of the one dimension data.

    window : str
        Window function such as rectwin, hamming (default).

    frame_overlap : float
        Length of each frame overlaps in second.

    frame_length : float
        Length of each frame in second.

    fs : float
        Sampling frequency.

    Returns
    -------
    s : numpy.ndarray
        Spectrogram of x.

    t : numpy.ndarray
        Value corresponds to the center of each frame (x-axis) in sec.

    f : numpy.ndarray
        Vector of frequencies (y-axis) in Hz.

    """

    assert x is not None and isinstance(x, xr.DataArray)

    if fs is None:
        dt = np.median(np.diff(x.time.data).astype(float)) * 1e-9
        fs = 1 / dt

    n_per_seg = np.round(frame_length * fs).astype(int)  # convert ms to points
    n_overlap = np.round(frame_overlap * fs).astype(int)  # convert ms to points

    f, t, s = signal.spectrogram(x, fs=fs, window=window, nperseg=n_per_seg, noverlap=n_overlap,
                                 mode='complex')

    return f, t, s



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
plot_spectr.py

@author : Louis RICHARD
"""

import seaborn as sns
import matplotlib.pyplot as plt
import matplotlib.dates as mdates

from matplotlib import colors
from pandas.plotting import register_matplotlib_converters
register_matplotlib_converters()

# plt.style.use("seaborn-whitegrid")
locator = mdates.AutoDateLocator(minticks=3, maxticks=7)
formatter = mdates.ConciseDateFormatter(locator)
# sns.set_context("paper")
# plt.rc('lines', linewidth=1)


def plot_spectr(ax=None, inp=None, yscale="", cscale="", clim=None, cmap="", cbar=True, **kwargs):
    """Plot a spectrogram using pcolormesh.

    Parameters
    ----------
    ax : axes
        Target axis to plot. If None create a new figure.

    inp : DataArray
        Input 2D data to plot.

    yscale : str
        Y-axis flag. Default is "" (linear).

    cscale : str
        C-axis flag. Default is "" (linear).

    clim : list
        C-axis bounds. Default is None (autolim).

    cmap : str
        Colormap. Default is jet.

    cbar : bool
        Flag for colorbar. Set to False to hide.

    Returns
    -------
    fig : figure
        to fill.

    axs : axes
        to fill.

    caxs : caxes
        Only if cbar is True.

    """

    if ax is None:
        fig, ax = plt.subplots(1)
    else:
        fig = plt.gcf()

    if not cmap:
        cmap = "jet"

    if cscale == "log":
        if clim is not None and isinstance(clim, list):
            options = dict(norm=colors.LogNorm(vmin=clim[0], vmax=clim[1]), cmap=cmap, shading='auto')
        else:
            options = dict(norm=colors.LogNorm(), cmap=cmap, shading='auto')
    else:
        if clim is not None and isinstance(clim, list):
            options = dict(cmap=cmap, vmin=clim[0], vmax=clim[1], shading='auto')
        else:
            options = dict(cmap=cmap, vmin=None, vmax=None, shading='auto')

    t, y = [inp.coords[inp.dims[0]], inp.coords[inp.dims[1]]]

    im = ax.pcolormesh(t, y, inp.data.T, **options)

    if yscale == "log":
        ax.set_yscale("log")

    if cbar:
        if "pad" in kwargs:
            pad = kwargs["pad"]
        else:
            pad = 0.01

        pos = ax.get_position()
        cax = fig.add_axes([pos.x0+pos.width+pad, pos.y0, 0.01, pos.height])
        fig.colorbar(mappable=im, cax=cax, ax=ax)

        return ax, cax
    else:
        return ax



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
copy_files.py

@author : Louis RICHARD
"""

import os

from .list_files import list_files
from .mms_config import CONFIG


def copy_files(var=None, tint=None, mms_id=1, target_dir="./data/"):
    """Copy files from NAS24 to the target path

    Parameters
    ----------
    var : dict
        Dictionary containing 4 keys
            * var["inst"] : name of the instrument
            * var["tmmode"] : data rate
            * var["lev"] : data level
            * var["dtype"] : data type

    tint : list of str
        Time interval

    mms_id : str or int
        Index of the spacecraft

    target_dir : str
        Target path

    """
    assert var is not None and isinstance(var, dict)
    assert tint is not None and isinstance(tint, list)
    assert isinstance(mms_id, int)
    assert isinstance(target_dir, str)

    mms_path = CONFIG["local_data_dir"] + "/"

    files = list_files(tint, mms_id, var)

    for file in files:
        relative_path = os.path.split(file)[0].replace(mms_path, "")
        path = os.path.join(target_dir, relative_path)

        if not os.path.exists(path):
            os.makedirs(path)
            os.popen('cp {} {}'.format(file, os.path.join(path, os.path.split(file)[1])))
        else:
            os.popen('cp {} {}'.format(file, os.path.join(path, os.path.split(file)[1])))

    return


#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
histogram.py

@author : Louis RICHARD
"""

import numpy as np
import xarray as xr


def histogram(inp=None, bins=100, normed=True):
    """
    Computes 1D histogram of the inp with bins bins

    Parameters
    ----------
    inp : xarray.DataArray
        Time series of the input scalar variable.

    bins : int
        Number of bins.

    normed : bool
        Normalize the PDF.

    Returns
    -------
    out : xarray.DataArray
        1D distribution of the input time series.

    """

    assert inp is not None and isinstance(inp, xr.DataArray)

    hist, bins = np.histogram(inp.data, bins=bins, normed=normed)
    bin_center = (bins[1:] + bins[:-1]) * 0.5

    out = xr.DataArray(hist, coords=[bin_center], dims=["bins"])

    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
get_feeps_alleyes.py

@author : Louis RICHARD
"""

import numpy as np
import xarray as xr

from .get_feeps_oneeye import get_feeps_oneeye
from .get_feeps_active_eyes import get_feeps_active_eyes


def get_feeps_alleyes(tar_var="fluxe_brst_l2", tint=None, mms_id=1, verbose=True):
    """Read energy spectrum of the selected specie in the selected energy range for all FEEPS eyes.

    Parameters
    ----------
    tar_var : str
        Key of the target variable like {data_unit}{specie}_{data_rate}_{data_lvl}.

    tint : list of str
        Time interval.

    mms_id : int or float or str
        Index of the spacecraft.

    verbose : bool
        Set to True to follow the loading. Default is True.

    Returns
    -------
    out : xarray.Dataset
        Dataset containing the energy spectrum of the available eyes of the Fly's Eye Energetic
        Particle Spectrometer.

    """

    assert isinstance(tar_var, str)
    assert tint is not None and isinstance(tint, list)
    assert isinstance(tint[0], str) and isinstance(tint[1], str)
    assert isinstance(mms_id, (int, str)) and int(mms_id) in np.arange(1, 5)
    assert isinstance(verbose, bool)

    mms_id = int(mms_id)

    specie = tar_var.split("_")[0][-1]

    var = {"tmmode": tar_var.split("_")[1], "lev": tar_var.split("_")[2], "mmsId": mms_id}

    if specie == "e":
        var["dtype"] = "electron"
    elif specie == "i":
        var["dtype"] = "ion"
    else:
        raise ValueError("Invalid specie")

    active_eyes = get_feeps_active_eyes(var, tint, mms_id)

    e_ids = [f"{k}-{s:d}" for k in active_eyes for s in active_eyes[k]]

    out_dict = {}

    for e_id in e_ids:
        out_dict[e_id] = get_feeps_oneeye(tar_var, e_id, tint, mms_id, verbose)

    out = xr.Dataset(out_dict)

    out.attrs = var

    out.attrs["species"] = "{}s".format(var["dtype"])

    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
plasma_beta.py

@author : Louis RICHARD
"""

import xarray as xr

from ..mms import rotate_tensor

from . import norm, resample, ts_scalar


def plasma_beta(b_xyz=None, p_xyz=None):
    """Computes plasma beta at magnetic field sampling

    .. math::

        \\beta = \\frac{P_{th}}{P_b}

    where : :math:`P_b = B^2 / 2 \\mu_0`

    Parameters
    ----------
    b_xyz : xarray.DataArray
        Time series of the magnetic field.

    p_xyz : xarray.DataArray
        Time series of the pressure tensor.

    Returns
    -------
    beta : xarray.DataArray
        Time series of the plasma beta at magnetic field sampling.

    """

    assert p_xyz is not None and isinstance(p_xyz, xr.DataArray)
    assert b_xyz is not None and isinstance(b_xyz, xr.DataArray)

    p_xyz = resample(p_xyz, b_xyz)

    p_fac = rotate_tensor(p_xyz, "fac", b_xyz, "pp")

    # Scalar temperature
    p_mag = (p_fac[0, 0] + (p_fac[1, 1] + p_fac[2, 2]) / 2) / 2

    # Magnitude of the magnetic field
    b_mag = norm(b_xyz)

    # Compute plasma beta
    beta = p_mag / (b_mag * 1e-5) ** 2

    time = b_xyz.time.data
    beta = ts_scalar(time, beta)

    return beta



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
find_closest.py

@author : Louis RICHARD
"""

import numpy as np

from scipy import interpolate


def find_closest(t1=None, t2=None):
    """Finds pairs that are closest to each other in two time series.

    Parameters
    ----------
    t1 : numpy.ndarray
        Vector with time instants.

    t2 : numpy.ndarray
        Vector with time instants.

    Returns
    -------
    t1new : numpy.ndarray
        Identified time instants that are closest each other.

    t2new : numpy.ndarray
        Identified time instants that are closest each other.

    ind1new : numpy.ndarray
        Identified time instants that are closest each other.

    ind2new : numpy.ndarray
        Identified time instants that are closest each other.

    """

    assert t1 is not None
    assert t2 is not None

    t1_orig = t1
    t2_orig = t2
    flag = True

    nt1, nt2 = [len(t) for t in [t1, t2]]

    while flag:
        flag_t1 = np.zeros(t1.shape)
        tckt1 = interpolate.interp1d(t1, np.arange(nt1), kind="nearest", fill_value="extrapolate")
        ind = tckt1(t2)
        flag_t1[ind] = 1

        flag_t2 = np.zeros(t2.shape)
        tckt2 = interpolate.interp1d(t2, np.arange(nt2), kind="nearest", fill_value="extrapolate")
        ind = tckt2(t1)
        flag_t2[ind] = 1

        ind_zeros_t1 = np.where(flag_t1 == 0)[0]
        ind_zeros_t2 = np.where(flag_t2 == 0)[0]
        if ind_zeros_t1:
            t1 = np.delete(t1, ind_zeros_t1)
        elif ind_zeros_t2:
            t2 = np.delete(t2, ind_zeros_t2)
        else:
            break

    t1new = t1
    t2new = t2

    tckt1_orig = interpolate.interp1d(t1_orig, np.arange(nt1), kind="nearest")
    ind1new = tckt1_orig(t1new)

    tckt2_orig = interpolate.interp1d(t2_orig, np.arange(nt2), kind="nearest")
    ind2new = tckt2_orig(t2new)

    return t1new, t2new, ind1new, ind2new



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
c_4_grad.py

@author : Louis RICHARD
"""

import numpy as np
import xarray as xr

from .resample import resample
from .c_4_k import c_4_k
from .normalize import normalize
from .avg_4sc import avg_4sc
from .dot import dot
from .cross import cross


def c_4_grad(r_list=None, b_list=None, method="grad"):
    """Calculate gradient of physical field using 4 spacecraft technique in [2]_ [3]_.

    Parameters
    ----------
    r_list : list of xarray.DataArray
        Time series of the positions of the spacecraft

    b_list : list of xarray.DataArray
        Time series of the magnetic field at the corresponding positions

    method : str
        Method flag :
            * "grad" : compute gradient (default)
            * "div" : compute divergence
            * "curl" : compute curl
            * "bdivb" : compute b.div(b)
            * "curv" : compute curvature

    Returns
    -------
    out : xarray.DataArray
        Time series of the derivative of the input field corresponding to the method

    See also
    --------
    pyrfu.pyrf.c_4_k : Calculates reciprocal vectors in barycentric coordinates.

    References
    ----------
    .. [2]	Dunlop, M. W., A. Balogh, K.-H. Glassmeier, and P. Robert (2002a), Four-point Cluster
            application of	magnetic field analysis tools: The Curl- ometer, J. Geophys. Res.,
            107(A11), 1384, doi : https://doi.org/10.1029/2001JA005088.

    .. [3]	Robert, P., et al. (1998), Accuracy of current determination, in Analysis Methods for
            Multi-Spacecraft Data, edited by G. Paschmann and P. W. Daly, pp. 395–418, Int. Space
            Sci. Inst., Bern. url : http://www.issibern.ch/forads/sr-001-16.pdf

    Examples
    --------
    >>> from pyrfu import mms, pyrf
    >>> # Time interval
    >>> tint = ["2019-09-14T07:54:00.000", "2019-09-14T08:11:00.000"]
    >>> # Load magnetic field and spacecraft position
    >>> b_mms = [mms.get_data("B_gse_fgm_srvy_l2", tint, ic) for ic in range(1, 5)]
    >>> r_mms = [mms.get_data("R_gse", tint, ic) for ic in range(1, 5)]
    >>> grad_b = pyrf.c_4_grad(r_mms, b_mms, "grad")

    """

    assert r_list is not None and isinstance(r_list, list) and len(r_list) == 4
    assert b_list is not None and isinstance(b_list, list) and len(b_list) == 4

    for i in range(4):
        if not isinstance(r_list[i], xr.DataArray):
            raise TypeError("Spacecraft position must be DataArray")

        if not isinstance(b_list[i], xr.DataArray):
            raise TypeError("Magnetic field must be DataArray")

    # Resample with respect to 1st spacecraft
    r_list = [resample(r, b_list[0]) for r in r_list]
    b_list = [resample(b, b_list[0]) for b in b_list]

    # Compute reciprocal vectors in barycentric coordinates (see c_4_k)
    k_list = c_4_k(r_list)

    # Magnetic field at the center of mass of the tetrahedron
    b_avg = avg_4sc(b_list)

    b_dict = {"1": b_list[0], "2": b_list[1], "3": b_list[2], "4": b_list[3]}
    k_dict = {"1": k_list[0], "2": k_list[1], "3": k_list[2], "4": k_list[3]}

    mms_list = b_dict.keys()

    # Gradient of scalar/vector
    if len(b_dict["1"].shape) == 1:
        grad_b = np.zeros((len(b_dict["1"]), 3))

        for mms_id in mms_list:
            grad_b += k_dict[mms_id].data * np.tile(b_dict[mms_id].data, (3, 1)).T

    else:
        grad_b = np.zeros((len(b_dict["1"]), 3, 3))

        for i in range(3):
            for j in range(3):
                for mms_id in mms_list:
                    grad_b[:, j, i] += k_dict[mms_id][:, i].data * b_dict[mms_id][:, j].data

    # Gradient
    if method.lower() == "grad":
        out_data = grad_b

    # Divergence
    elif method.lower() == "div":
        div_b = np.zeros(len(b_dict["1"]))

        for mms_id in mms_list:
            div_b += dot(k_dict[mms_id], b_dict[mms_id]).data

        out_data = div_b

    # Curl
    elif method.lower() == "curl":
        curl_b = np.zeros((len(b_dict["1"]), 3))

        for mms_id in mms_list:
            curl_b += cross(k_dict[mms_id], b_dict[mms_id]).data

        out_data = curl_b

    # B.div(B)
    elif method.lower() == "bdivb":
        b_div_b = np.zeros(b_avg.shape)

        for i in range(3):
            b_div_b[:, i] = np.sum(b_avg.data * grad_b[:, i, :], axis=1)

        out_data = b_div_b

    # Curvature
    elif method.lower() == "curv":
        b_hat_list = [normalize(b) for b in b_list]

        curv = c_4_grad(r_list, b_hat_list, method="bdivb")

        out_data = curv.data

    else:
        raise ValueError("Invalid method")

    if len(out_data.shape) == 1:
        out = xr.DataArray(out_data, coords=[b_dict["1"].time], dims=["time"])

    elif len(out_data.shape) == 2:
        out = xr.DataArray(out_data, coords=[b_dict["1"].time, ["x", "y", "z"]],
                           dims=["time", "comp"])

    elif len(out_data.shape) == 3:
        out = xr.DataArray(
            out_data, coords=[b_dict["1"].time, ["x", "y", "z"], ["x", "y", "z"]],
            dims=["time", "vcomp", "hcomp"])

    else:
        raise TypeError("Invalid type")

    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
poynting_flux.py

@author : Louis RICHARD
"""

import numpy as np
import xarray as xr

from astropy.time import Time

from . import calc_fs, cross, dot, normalize, resample, time_clip


def poynting_flux(e_xyz=None, b_xyz=None, b0=None):
    """
    Estimates Poynting flux at electric field sampling as

    .. math::

        \\mathbf{S} = \\frac{\\mathbf{E}\\times\\mathbf{B}}{\\mu_0}

    if `b0` is given project the Poynting flux along `b0`


    Parameters
    ----------
    e_xyz : xarray.DataArray
        Time series of the electric field.

    b_xyz : xarray.DataArray
        Time series of the magnetic field.

    b0 : xarray.DataArray, optional
        Time series of the direction to project the Pointing flux.

    Returns
    -------
    s : xarray.DataArray
        Time series of the Pointing flux.

    s_z : xarray.DataArray
        Time series of the projection of the Pointing flux (only if b0).

    int_s : xarray.DataArray
        Time series of the time integral of the Pointing flux (if b0 integral along b0).

    """

    assert e_xyz is not None and isinstance(e_xyz, xr.DataArray)
    assert b_xyz is not None and isinstance(b_xyz, xr.DataArray)

    # check which Poynting flux to calculate
    flag_s_z, flag_int_s_z, flag_int_s = [False, False, False]

    if b0 is None:
        flag_int_s = True
    else:
        flag_s_z, flag_int_s_z = [True, True]

    # resample if necessary
    fs_b, fs_e = [calc_fs(b_xyz), calc_fs(e_xyz)]

    # interval where both E & B exist
    tmin = Time(max([min(e_xyz.time.data), min(b_xyz.time.data)]), format="datetime64").iso
    tmax = Time(min([max(e_xyz.time.data), max(b_xyz.time.data)]), format="datetime64").iso
    tint = [tmin, tmax]
    ee, bb = [time_clip(e_xyz, tint), time_clip(b_xyz, tint)]
    
    if fs_e < fs_b:
        e = resample(ee, bb)
        b = bb
        fs = fs_b
    elif fs_e > fs_b:
        b = resample(bb, ee)
        e = ee
        fs = fs_e
    else:
        e = ee
        b = bb
        print("assuming the same sampling. Interpolating B and E to 2x E sampling.")

    """
    else
      disp('assuming the same sampling. Interpolating B and E to 2x E sampling.');
      t=sort([ee(:,1);ee(:,1)+0.5/fs_e])
      e=irf_resamp(ee,t)
      b=irf_resamp(bb,t);Fs=2*fs_e;
    end
    """

    # Calculate Poynting flux
    s = cross(e, b) / (4 * np.pi / 1e7) * 1e-9

    if flag_s_z:
        b_m = resample(b0, e)
        s_z = dot(normalize(b_m), s)

    # time integral of Poynting flux along ambient magnetic field
    if flag_int_s_z:
        ss_z = s_z

        idx = np.isnan(s_z.data)

        ss_z[idx] = 0  # set to zero points where Sz=NaN

        int_s_z = np.cumsum(ss_z) / fs
        return s, s_z, int_s_z

    if flag_int_s:  # time integral of all Poynting flux components
        ss = s
        idx = np.isnan(s[:, 2].data)
        ss[idx] = 0  # set to zero points where Sz=NaN

        int_s = np.cumsum(ss) / fs
        return s, int_s



# !/usr/bin/env python
# -*- coding: utf-8 -*-
"""
colorbar.py

@author : Louis RICHARD
"""

import matplotlib.pyplot as plt

from pandas.plotting import register_matplotlib_converters
register_matplotlib_converters()


def colorbar(im, ax, pad=0.01):
    """Add colorbar to ax corresponding to im.

    Parameters
    ----------
    im : to fill
        to fill.

    ax : matplotlib.pyplot.subplotsaxes
        Axis of plot.

    pad : float, optional


    Returns
    -------
    cax : coloraxis

    """
    pos = ax.get_position()
    fig = plt.gcf()
    cax = fig.add_axes([pos.x0 + pos.width + pad, pos.y0, pad, pos.height])
    fig.colorbar(im, cax)

    return cax



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
extend_tint.py

@author : Louis RICHARD
"""

from dateutil.parser import parse as date_parse
from astropy.time import Time


def extend_tint(tint=None, ext=None):
    """Extends time interval.

    Parameters
    ----------
    tint : list of str
        Reference time interval to extend.

    ext : list of float or int
        Number of seconds to extend time interval [left extend, right extend].

    Returns
    -------
    tint : list of str
        Extended time interval.

    Examples
    --------
    >>> from pyrfu import pyrf
    >>> # Time interval
    >>> tint = ["2015-10-30T05:15:42.000", "2015-10-30T05:15:54.000"]
    >>> # Spacecraft index
    >>> ic = 3
    >>> # Load spacecraft position
    >>> tint_long = pyrf.extend_tint(tint, [-100, 100])

    """

    if ext is None:
        ext = [-60, 60]

    # Convert to unix format
    start_time, end_time = [Time(date_parse(tint_bound), format="datetime").unix for tint_bound in
                            tint]

    # extend interval
    start_time, end_time = [start_time + ext[0], end_time + ext[1]]

    # back to iso format
    start_time, end_time = [Time(bound, format="unix").iso for bound in [start_time, end_time]]

    tint_long = [start_time, end_time]

    return tint_long



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
new_xyz.py

@author : Louis RICHARD
"""

import xarray as xr
import numpy as np


def new_xyz(inp=None, trans_mat=None):
    """Transform the input field to the new frame.

    Parameters
    ----------
    inp : xarray.DataArray
        Time series of the input field in the original coordinate system.

    trans_mat : numpy.ndarray
        Transformation matrix.

    Returns
    -------
    out : xarray.DataArray
        Time series of the input in the new frame.

    Examples
    --------
    >>> from pyrfu import mms, pyrf
    >>> # Time interval
    >>> tint = ["2019-09-14T07:54:00.000", "2019-09-14T08:11:00.000"]
    >>> # Spacecraft indices
    >>> mms_id = 1
    >>> # Load magnetic field and electric field
    >>> b_xyz = mms.get_data("B_gse_fgm_srvy_l2", tint, mms_id)
    >>> e_xyz = mms.get_data("E_gse_edp_fast_l2", tint, mms_id)
    >>> # Compute MVA frame
    >>> b_lmn, l, mva = pyrf.mva(b_xyz)
    >>> # Move electric field to the MVA frame
    >>> e_lmn = pyrf.new_xyz(e_xyz, mva)

    """

    assert inp is not None and isinstance(inp, xr.DataArray)
    assert trans_mat is not None and isinstance(trans_mat, np.ndarray)

    if inp.data.ndim == 3:
        out_data = np.matmul(np.matmul(trans_mat.T, inp.data), trans_mat)
    else:
        out_data = (trans_mat.T @ inp.data.T).T

    out = xr.DataArray(out_data, coords=inp.coords, dims=inp.dims, attrs=inp.attrs)

    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
get_ts.py

@author : Louis RICHARD
"""

import bisect
import numpy as np
import xarray as xr

from spacepy import pycdf
from dateutil import parser


def get_ts(file_path="", cdf_name="", tint=None):
    """
    Read field named cdf_name in file and convert to time series

    Parameters
    ----------
    file_path : str
        Path of the cdf file

    cdf_name : str
        Name of the target variable in the cdf file

    tint : list
        Time interval

    Returns
    -------
    out : xarray.DataArray
        Time series of the target variable in the selected time interval

	TODO : remove th exec

    """

    assert isinstance(file_path, str) and file_path
    assert isinstance(cdf_name, str) and cdf_name
    assert tint is not None and isinstance(tint, list)
    assert isinstance(tint[0], str) and isinstance(tint[1], str)

    x, y, z, w = [{}, {}, {}, {}]

    out_dict = {}

    with pycdf.CDF(file_path) as f:
        depend0_key = f[cdf_name].attrs["DEPEND_0"]
        start_ind, stop_ind = [bisect.bisect_left(f[depend0_key], parser.parse(tint[0])),
                               bisect.bisect_left(f[depend0_key], parser.parse(tint[1]))]

        x["data"], x["attrs"] = [f[depend0_key][start_ind:stop_ind], {}]

        for k in f[depend0_key].attrs.keys():
            x["attrs"][k] = f[depend0_key].attrs[k]
            if isinstance(x["attrs"][k], str) and x["attrs"][k] in f.keys() and not k == "LABLAXIS":
                try:
                    # If array
                    x["attrs"][k] = f[x["attrs"][k]][start_ind:stop_ind, ...]
                except IndexError:
                    # If scalar
                    x["attrs"][k] = f[x["attrs"][k]][...]

        if "DEPEND_1" in f[cdf_name].attrs or "REPRESENTATION_1" in f[cdf_name].attrs:
            try:
                depend1_key = f[cdf_name].attrs["DEPEND_1"]
            except KeyError:
                depend1_key = f[cdf_name].attrs["REPRESENTATION_1"]

            if depend1_key == "x,y,z":
                y["data"], y["attrs"] = [np.array(depend1_key.split(",")), {"LABLAXIS": "comp"}]
            else:
                try:
                    y["data"] = f[depend1_key][start_ind:stop_ind, :]
                except IndexError:
                    y["data"] = f[depend1_key][...]

                # If vector components remove magnitude index

                if len(y["data"]) == 4 and all(y["data"] == ["x", "y", "z", "r"]):
                    y["data"] = y["data"][:-1]
                # if y is 2d get only first row assuming that the bins are the same
                elif y["data"].ndim == 2:
                    try:
                        y["data"] = y["data"][0, :]
                    except IndexError:
                        pass

                y["attrs"] = {}

                # Get attributes
                for k in f[depend1_key].attrs.keys():
                    y["attrs"][k] = f[depend1_key].attrs[k]

                    if isinstance(y["attrs"][k], str) and y["attrs"][k] in f.keys():
                        if k not in ["DEPEND_0", "LABLAXIS"]:
                            try:
                                y["attrs"][k] = f[y["attrs"][k]][start_ind:stop_ind, ...]
                            except:
                                y["attrs"][k] = f[y["attrs"][k]][...]
                            # If attrs is 2D get only first row
                            if y["attrs"][k].ndim == 2:
                                try:
                                    y["attrs"][k] = y["attrs"][k][0, :]
                                except IndexError:
                                    pass

                # Remove spaces in label
                try:
                    y["attrs"]["LABLAXIS"] = y["attrs"]["LABLAXIS"].replace(" ", "_")

                    if y["attrs"]["LABLAXIS"] == "Diffential_energy_channels":
                        y["attrs"]["LABLAXIS"] = "Differential_energy_channels"

                except KeyError:
                    y["attrs"]["LABLAXIS"] = "comp"

        elif "afg" in cdf_name or "dfg" in cdf_name:
            y["data"] = ["x", "y", "z"]
            y["attrs"] = {"LABLAXIS": "comp"}

        if "DEPEND_2" in f[cdf_name].attrs or "REPRESENTATION_2" in f[cdf_name].attrs:
            try:
                depend2_key = f[cdf_name].attrs["DEPEND_2"]
            except KeyError:
                depend2_key = f[cdf_name].attrs["REPRESENTATION_2"]

            if depend2_key == "x,y,z":
                z["data"] = np.array(depend2_key.split(","))

                z["attrs"] = {"LABLAXIS": "comp"}
            else:
                z["data"] = f[depend2_key][...]

                z["attrs"] = {}

                for k in f[depend2_key].attrs.keys():
                    z["attrs"][k] = f[depend2_key].attrs[k]

                    if isinstance(z["attrs"][k], str) and z["attrs"][k] in f.keys() and \
                            not k == "DEPEND_0":
                        z["attrs"][k] = f[z["attrs"][k]][start_ind:stop_ind, ...]

                if "LABLAXIS" not in z["attrs"].keys():
                    z["attrs"]["LABLAXIS"] = "comp"

        if "DEPEND_3" in f[cdf_name].attrs or "REPRESENTATION_3" in f[cdf_name].attrs and \
                f[cdf_name].attrs["REPRESENTATION_3"] != "x,y,z":

            try:
                depend3_key = f[cdf_name].attrs["DEPEND_3"]
            except KeyError:
                depend3_key = f[cdf_name].attrs["REPRESENTATION_3"]

            w["data"] = f[depend3_key][...]

            if w["data"].ndim == 2:
                try:
                    w["data"] = w["data"][0, :]
                except IndexError:
                    pass

            w["attrs"] = {}
            for k in f[depend3_key].attrs.keys():
                w["attrs"][k] = f[depend3_key].attrs[k]

                if isinstance(w["attrs"][k], str) and w["attrs"][k] in f.keys() and \
                        not k == "DEPEND_0":
                    w["attrs"][k] = f[w["attrs"][k]][start_ind:stop_ind, ...]

            if "LABLAXIS" not in w["attrs"].keys():
                w["attrs"]["LABLAXIS"] = "comp"

        if "sector_mask" in cdf_name:
            y["data"] = f[f[cdf_name.replace("sector_mask", "intensity")].attrs["DEPEND_1"]][...]

            y["attrs"] = {}

            y_attrs_keys = f[
                f[cdf_name.replace("sector_mask", "intensity")].attrs["DEPEND_1"]].attrs.keys()

            for k in y_attrs_keys:
                y["attrs"][k] = \
                f[f[cdf_name.replace("sector_mask", "intensity")].attrs["DEPEND_1"]].attrs[k]

            y["attrs"]["LABLAXIS"] = y["attrs"]["LABLAXIS"].replace(" ", "_")

        if "edp_dce_sensor" in cdf_name:
            y["data"] = ["x", "y", "z"]

            y["attrs"] = {"LABLAXIS": "comp"}

        out_dict["data"] = f[cdf_name][start_ind:stop_ind, ...]

        if out_dict["data"].ndim == 2 and out_dict["data"].shape[1] == 4:
            out_dict["data"] = out_dict["data"][:, :-1]

        out_dict["attrs"] = {}

        for k in f[cdf_name].attrs:
            out_dict["attrs"][k] = f[cdf_name].attrs[k]

    if x and not y and not z and not w:
        dims, coordinates = [["time"], [x["data"]]]

        out = xr.DataArray(out_dict["data"], coords=coordinates, dims=dims, attrs=out_dict["attrs"])
        exec("out."+dims[0]+".attrs = x['attrs']")

    elif x and y and not z and not w:
        dims, coordinates = [["time", y["attrs"]["LABLAXIS"]], [x["data"], y["data"]]]

        out = xr.DataArray(out_dict["data"], coords=coordinates, dims=dims, attrs=out_dict["attrs"])
        exec("out."+dims[0]+".attrs = x['attrs']")
        exec("out."+dims[1]+".attrs = y['attrs']")

    elif x and y and z and not w:

        if y["attrs"]["LABLAXIS"] == z["attrs"]["LABLAXIS"]:
            y["attrs"]["LABLAXIS"] = "rcomp"
            z["attrs"]["LABLAXIS"] = "ccomp"

        dims = ["time", y["attrs"]["LABLAXIS"], z["attrs"]["LABLAXIS"]]
        coordinates = [x["data"], y["data"], z["data"]]

        out = xr.DataArray(out_dict["data"], coords=coordinates, dims=dims, attrs=out_dict["attrs"])
        exec("out."+dims[0]+".attrs = x['attrs']")
        exec("out."+dims[1]+".attrs = y['attrs']")
        exec("out."+dims[2]+".attrs = z['attrs']")
    elif x and y and z and w:
        if z["attrs"]["LABLAXIS"] == w["attrs"]["LABLAXIS"]:
            z["attrs"]["LABLAXIS"] = "rcomp"
            w["attrs"]["LABLAXIS"] = "ccomp"

        dims = ["time", y["attrs"]["LABLAXIS"], z["attrs"]["LABLAXIS"], w["attrs"]["LABLAXIS"]]
        coordinates = [x["data"], y["data"], z["data"], w["data"]]

        out = xr.DataArray(out_dict["data"], coords=coordinates, dims=dims, attrs=out_dict["attrs"])
        exec("out."+dims[0]+".attrs = x['attrs']")
        exec("out."+dims[1]+".attrs = y['attrs']")
        exec("out."+dims[2]+".attrs = z['attrs']")
        exec("out."+dims[3]+".attrs = w['attrs']")

    else:
        raise NotImplementedError

    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
solid_angle.py

@author : Louis RICHARD
"""

import numpy as np


def solid_angle(x=None, y=None, z=None):
    """Calculates the solid angle of three vectors making up a triangle in a unit sphere with the
    sign taken into account.

    Parameters
    ----------
    x : numpy.ndarray
        First vector.

    y : numpy.ndarray
        Second vector.

    z : numpy.ndarray
        Third vector.

    Returns
    -------
    angle : float
        Solid angle.

    """

    assert x is not None and isinstance(x, np.ndarray) and x.ndim == 1 and len(x) == 3
    assert y is not None and isinstance(y, np.ndarray) and y.ndim == 1 and len(y) == 3
    assert z is not None and isinstance(z, np.ndarray) and z.ndim == 1 and len(z) == 3

    # Calculate the smaller angles between the vectors around origin
    a = np.arccos(np.sum(z * y))
    b = np.arccos(np.sum(x * z))
    c = np.arccos(np.sum(y * x))

    # Calculate the angles in the spherical triangle (Law of Cosines)
    u = np.arccos((np.cos(a) - np.cos(b) * np.cos(c)) / (np.sin(b) * np.sin(c)))
    v = np.arccos((np.cos(b) - np.cos(a) * np.cos(c)) / (np.sin(a) * np.sin(c)))
    w = np.arccos((np.cos(c) - np.cos(b) * np.cos(a)) / (np.sin(b) * np.sin(a)))

    # Calculates the Surface area on the unit sphere (solid angle)
    angle = (u + v + w - np.pi)
    # Calculate the sign of the area
    var = np.cross(z, y)
    div = np.sum(var * x)
    sgn = np.sign(div)

    # Solid angle with sign taken into account
    angle = sgn * angle

    return angle



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
edb.py

@author : Louis RICHARD
"""

import numpy as np

from .resample import resample
from .ts_scalar import ts_scalar
from .ts_vec_xyz import ts_vec_xyz


def edb(e=None, b0=None, angle_lim=20, flag_method="E.B=0"):
    """Compute Ez under assumption :math:`\\mathbf{E}.\\mathbf{B}=0` or
    :math:`\\mathbf{E}.\\mathbf{B} \\approx 0`

    Parameters
    ----------
    e : xarray.DataArray
        Time series of the electric field.

    b0 : xarray.DataArray
        Time series of the background magnetic field.

    flag_method : str
        Assumption on the direction of the measured electric field :
            "e.b=0" :  :math:`\\mathbf{E}.\\mathbf{B}=0`.
            "e_par" :  :math:`\\mathbf{E}` field along the B projection is coming from parallel
            electric field.
            "e_perp+nan" : to fill.

    angle_lim : float
        B angle with respect to the spin plane should be less than angle_lim degrees otherwise Ez
        is set to 0.

    Returns
    -------
    ed : xarray.DataArray
        Time series of the electric field output.

    d : xarray.DataArray
        Time series of the B elevation angle above spin plane.

    Examples
    --------
    >>> from pyrfu import mms, pyrf
    >>> # Time interval
    >>> tint = ["2019-09-14T07:54:00.000","2019-09-14T08:11:00.000"]
    >>> # Spacecraft indices
    >>> mms_id = 1
    >>> # Load magnetic field, electric field and spacecraft position
    >>> b_xyz = mms.get_data("B_gse_fgm_srvy_l2", tint, mms_id)
    >>> e_xyz = mms.get_data("E_gse_edp_fast_l2", tint, mms_id)
    >>> # Compute Ez
    >>> e_z, alpha = pyrf.edb(e_xyz, b_xyz)

    """

    if e is None:
        raise ValueError("edb requires at least two inputs")
    if b0 is None:
        raise ValueError("edb requires at least two inputs")

    default_value = 0
    if flag_method.lower() == "e_perp+nan":
        default_value = np.nan

        flag_method = "e.b=0"

    if len(b0) != len(e):
        b0 = resample(b0, e)

    bd = b0.data
    ed = e.data
    ed[:, -1] *= default_value

    if flag_method.lower() == "e.b=0":
        # Calculate using assumption E.B=0
        d = np.arctan2(bd[:, 2], np.sqrt(bd[:, 0] ** 2 + bd[:, 1] ** 2)) * 180 / np.pi

        ind = np.abs(d) > angle_lim

        if True in ind:
            ed[ind, 2] = -(ed[ind, 0] * bd[ind, 0] + ed[ind, 1] * bd[ind, 1]) / bd[ind, 2]

    elif flag_method.lower() == "e_par":
        # Calculate using assumption that E field along the B projection is coming from parallel
        # electric field
        d = np.arctan2(bd[:, 2], np.sqrt(bd[:, 0] ** 2 + bd[:, 1] ** 2)) * 180 / np.pi

        ind = np.abs(d) < angle_lim

        if True in ind:
            ed[ind, 2] = (ed[ind, 0] * bd[ind, 0] + ed[ind, 1] * bd[ind, 1])
            ed[ind, 2] = ed[ind, 2] * bd[ind, 2] / (bd[ind, 0] ** 2 + bd[ind, 1] ** 2)

    else:
        raise ValueError("Invalid flag")

    ed = ts_vec_xyz(e.time.data, ed, {"UNITS": e.attrs["UNITS"]})
    d =  ts_scalar(e.time.data, d, {"UNITS": "degrees"})

    return ed, d



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
integrate.py

@author : Louis RICHARD
"""

import numpy as np
import xarray as xr
from astropy.time import Time


def integrate(inp=None, time_step=None):
    """Integrate time series.

    Parameters
    ----------
    inp : xarray.DataArray
        Time series of the variable to integrate.

    time_step : float
        Time steps threshold. All time_steps larger than 3*time_step are assumed data gaps,
        default is that time_step is the smallest value of all time_steps of the time series.

    Returns
    -------
    out : xarray.DataArray
        Time series of the time integrated input.

    Examples
    --------
    >>> from pyrfu import mms, pyrf
    >>> # Time interval
    >>> tint = ["2015-12-14T01:17:40.200", "2015-12-14T01:17:41.500"]
    >>> # Spacecraft index
    >>> mms_id = 1
    >>> # Load magnetic field and electric field
    >>> b_xyz = mms.get_data("B_gse_fgm_brst_l2", tint, mms_id)
    >>> e_xyz = mms.get_data("E_gse_edp_brst_l2", tint, mms_id)
    >>> # Convert electric field to field aligned coordinates
    >>> e_xyzfac = pyrf.convert_fac(e_xyz, b_xyz, [1, 0, 0])

    """

    if inp is None:
        raise ValueError("integrate requires at least one argument")

    if not isinstance(inp, xr.DataArray):
        raise TypeError("inp must be a DataArray")

    time_tmp = Time(inp.time.data, format="datetime64").unix
    data_tmp = inp.data
    unit_tmp = inp.attrs["UNITS"]

    x = np.hstack([time_tmp, data_tmp])

    dt = np.hstack([0, np.diff(x[:, 0])])

    if time_step is None:
        time_steps = np.diff(x[:, 0])

        ind_min = np.argmin(time_steps)

        # remove the smallest time step in case some problems
        time_steps = np.delete(time_steps, ind_min)

        time_step = np.min(time_steps)

    dt[dt > 3 * time_step] = 0

    x_int = x
    for j in range(1, x_int.shape[1]):
        j_ok = ~np.isnan(x_int[:, j])

        x_int[j_ok, j] = np.cumsum(x[j_ok, j] * dt[j_ok])

    out = xr.DataArray(x[:, 1:], coords=inp.coords, dims=inp.dims)
    out.attrs["UNITS"] = unit_tmp + "*s"

    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
remove_repeated_points.py

@author : Louis RICHARD
"""

import numpy as np
import xarray as xr

from . import ts_vec_xyz, ts_scalar


def remove_repeated_points(inp=None):
    """Remove repeated elements in DataArray or structure data. Important when using defatt
    products. Must have a time variable.

    Parameters
    ----------
    inp : xarray.DataArray or dict
        Time series of the input variable.

    Returns
    -------
    out: xarray.DataArray or dict
        Time series of the cleaned input variable.

    """

    assert inp is not None and isinstance(inp, (xr.DataArray, dict))

    threshold = 100  # Points separated in time by less than 100ns are treated as repeats

    if isinstance(inp, xr.DataArray):
        diffs = np.diff(inp.time.data.view("i8") * 1e-9)

        no_repeat = np.ones(len(inp))
        no_repeat[diffs < threshold] = 0

        new_time, new_inp = [inp.time.data[no_repeat == 1], inp.data[no_repeat == 1, :]]

        if new_inp.ndim == 1:
            new_data = ts_scalar(new_time, new_inp)
        elif new_inp.ndim == 2:
            new_data = ts_vec_xyz(new_time, new_inp)
        elif new_inp.ndim == 3:
            new_data = ts_vec_xyz(new_time, new_inp)
        else:
            raise TypeError("Invalid data dimension")

    elif isinstance(inp, dict) and ("time" in inp):
        if inp["time"].dtype == "<M8[ns]":
            diffs = np.diff(inp["time"].view("i8") * 1e-9)
        else:
            diffs = np.diff(inp["time"])

        no_repeat = np.ones(len(inp["time"]))
        no_repeat[diffs < threshold] = 0

        var_names = inp.keys()

        for var_name in var_names:
            inp[var_name] = inp[var_name][no_repeat == 1, :]

        new_data = inp
    else:
        new_data = inp  # no change to input if it's not a DataArray or structure

    return new_data



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
test_pyrf.py

@author : Louis RICHARD
"""


from pyrfu import mms, pyrf

import unittest

import numpy as np


class TestPyrf(unittest.TestCase):
    def setUp(self):
        """integration test setup."""
        tint = ["2019-09-14T08:00:00.000", "2019-09-14T08:00:30.000"]
        self.b_xyz = mms.get_data("B_gse_fgm_brst_l2", tint, 2)
        self.e_xyz = mms.get_data("E_gse_edp_brst_l2", tint, 2)

    def test_calc_fs(self):
        """integration test on sampling frequency computation"""
        fs = pyrf.calc_fs(self.b_xyz)

        self.assertEqual(np.round(fs), 128.0)

    def test_calc_dt(self):
        """integration test on sampling tme step computation"""
        dt = pyrf.calc_dt(self.b_xyz)

        self.assertEqual(int(dt * 1e9), 7813000)

    def test_cross(self):
        """integration test on vector cross product computation"""
        exb = pyrf.cross(self.e_xyz, self.b_xyz)
        res = pyrf.dot(exb, self.e_xyz)
        res = np.mean(res)
        self.assertTrue(res < 1e-5)

    def test_resample(self):
        """integration test on time resampling"""
        e_xyz = pyrf.resample(self.e_xyz, self.b_xyz)

        self.assertTrue((pyrf.resample(e_xyz, self.b_xyz).time.data == self.b_xyz.time.data).all())


if __name__ == "__main__":
    unittest.main()



from .plot_line import plot_line
from .plot_spectr import plot_spectr
from .mms_pl_config import mms_pl_config
from .pl_scatter_matrix import pl_scatter_matrix
from .zoom import zoom
from .colorbar import colorbar



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
probe_align_times.py

@author : Louis RICHARD
"""
import numpy as np
import xarray as xr
import matplotlib.pyplot as plt

from astropy.time import Time

from ..pyrf import ts_scalar, resample, extend_tint, time_clip, convert_fac


def probe_align_times(e_xyz=None, b_xyz=None, sc_pot=None, z_phase=None, plot_fig=False):
    """Returns times when field-aligned electrostatic waves can be characterized using
    interferometry techniques. The same alignment conditions as Graham et al., JGR, 2015 are
    used. Optional figure produced  showing E_FAC, probe fields, and probe potentials to view
    time delays between electric fields aligned with B.  Currently p5-p6 are not used in this
    routine; the analysis is the same as the one used for Cluster.

    For the figure the panels are :
        * (a) B in DMPA Coordinates
        * (b) Magnitude of B in and out of the spin plane
        * (c) Angles between B and probes 1 and 3 in the spin plane (angle between 0 and 90 degrees)
        * (d) Spacecraft potential from probes perpendicular to B
        * (e) E fields from p1-p4 and SC for probes closely aligned with B
        * (f) E in field-aligned coordinates
        * (g) E from probes p1-p2 and p3-p4.


    Parameters
    ----------
    e_xyz : xarray.DataArray
        Electric field in DSL coordinates, brst mode.

    b_xyz : xarray.DataArray
        Magnetic field in DMPA coordinates.

    sc_pot : xarray.DataArray
        L2 Spacecraft potential data. Timing corrections are applied in this

    z_phase : xarray.DataArray
        Spacecraft phase (z_phase). Obtained from ancillary_defatt.

    plot_fig : bool, optional
        Flag to plot fields and potentials which satisfy alignment conditions. Default is False

    Returns
    -------
    start_time1 : to fill
        Start times of intervals which satisfy the probe alignment conditions for probe combinates
        p1-p2.

    end_time1 : to fill
        End times of intervals which satisfy the probe alignment conditions for probe combinates
        p1-p2.

    start_time3 : to fill
        Start times of intervals which satisfy the probe alignment conditions for probe combinates
        p3-p4.

    end_time3 : to fill
        End times of intervals which satisfy the probe alignment conditions for probe combinates
        p3-p4.

    """

    # Correct for timing in spacecraft potential data.
    e12 = ts_scalar(sc_pot.time.data, (sc_pot.data[:, 0] - sc_pot.data[:, 1]) / .120)
    e34 = ts_scalar(sc_pot.time.data, (sc_pot.data[:, 2] - sc_pot.data[:, 3]) / .120)
    e56 = ts_scalar(sc_pot.time.data, (sc_pot.data[:, 4] - sc_pot.data[:, 4]) / .0292)

    v1 = ts_scalar(sc_pot.time.data, sc_pot.data[:, 0])
    v3 = ts_scalar(sc_pot.time.data + np.timedelta64(7629, "ns"), sc_pot.data[:, 2])
    v5 = ts_scalar(sc_pot.time.data + np.timedelta64(15259, "ns"), sc_pot.data[:, 4])

    e12.time.data += np.timedelta64(26703, "ns")
    e34.time.data += np.timedelta64(30518, "ns")
    e56.time.data += np.timedelta64(34332, "ns")

    v1, v3, v5 = [resample(v, v1) for v in [v1, v3, v5]]
    e12, e34, e56 = [resample(e, v1) for e in [e12, e34, e56]]

    v2 = v1 - e12 * 0.120
    v4 = v3 - e34 * 0.120
    v6 = v5 - e56 * 0.0292

    sc_pot = np.hstack([v1.data, v2.data, v3.data, v4.data, v5.data, v6.data])

    sc_pot = xr.DataArray(sc_pot, coords=[v1.time.data, np.arange(1, 7)], dims=["time", "probe"])

    t_limit = list(Time(sc_pot.time.data[[0, -1]], format="datetime64").isot)

    t_limit_long = extend_tint(t_limit, [-10, 10])

    b_xyz = time_clip(b_xyz, t_limit_long)
    b_xyz = resample(b_xyz, sc_pot)
    e_xyz = resample(e_xyz, sc_pot)

    z_phase = time_clip(z_phase, t_limit_long)

    # Remove repeated z_phase elements
    n_ph = len(z_phase)
    no_repeat = np.ones(n_ph)

    for i in range(1, n_ph):
        if z_phase.time.data[i] > z_phase.time.data[i - 1]:
            if z_phase.data[i] < z_phase.data[i - 1]:
                z_phase.data[i:] += 360.
        else:
            no_repeat[i] = 0

    z_phase_time = z_phase.time[no_repeat == 1]
    z_phase_data = z_phase.data[no_repeat == 1]

    z_phase = ts_scalar(z_phase_time, z_phase_data)
    z_phase = resample(z_phase, sc_pot)

    # Perform rotation on e_xyz into field - aligned coordinates
    e_fac = convert_fac(e_xyz, b_xyz, [1, 0, 0])

    # Probe angles in DSL or whatever
    idx_a, idx_b = [[1, 7, 2, 5], [6, 6, 3, 3]]
    phase_p = [z_phase.data / 180 * np.pi + i * np.pi / j for i, j in zip(idx_a, idx_b)]

    rp = [np.array([60 * np.cos(phase), 60 * np.sin(phase)]) for phase in phase_p]

    # Calculate angles between probes and direction of B in the spin plane.
    theta_pb = [None, None, None, None]

    for i in [0, 2]:
        theta_pb[i] = rp[i][:, 0] * b_xyz.data[:, 0] + rp[i][:, 1] * b_xyz.data[:, 1]
        theta_pb[i] /= np.sqrt(rp[i][:, 0] ** 2 + rp[i][:, 1] ** 2)
        theta_pb[i] /= np.sqrt(b_xyz[:, 0] ** 2 + b_xyz[:, 1] ** 2)
        theta_pb[i] = np.arccos(abs(theta_pb[i])) * 180 / np.pi

    theta_pb[1] = theta_pb[0]
    theta_pb[3] = theta_pb[2]

    sc_v12 = (sc_pot.data[:, 0] + sc_pot.data[:, 1]) / 2
    sc_v34 = (sc_pot.data[:, 2] + sc_pot.data[:, 3]) / 2

    e1, e3 = [(sc_pot.data[:, i] - sc_v) * 1e3 / 60 for i, sc_v in zip([0, 2], [sc_v34, sc_v12])]
    e2, e4 = [(sc_v - sc_pot.data[:, i]) * 1e3 / 60 for i, sc_v in zip([0, 2], [sc_v34, sc_v12])]

    es = [e1, e2, e3, e4]

    e12 = (sc_pot.data[:, 0] - sc_pot.data[:, 1]) * 1e3 / 120
    e34 = (sc_pot.data[:, 2] - sc_pot.data[:, 3]) * 1e3 / 120

    idx_b = np.sqrt(b_xyz.data[:, 0] ** 2 + b_xyz.data[:, 1] ** 2) < abs(b_xyz.data[:, 2])
    thresh_ang = 25.0

    for e, theta in zip(es, theta_pb):
        e[theta > thresh_ang] = np.nan
        e[idx_b] = np.nan

    sc_v12[theta_pb[2] > thresh_ang] = np.nan
    sc_v34[theta_pb[0] > thresh_ang] = np.nan

    sc_v12[idx_b] = np.nan
    sc_v34[idx_b] = np.nan

    if plot_fig:
        f, ax = plt.subplots(7, sharex="all", figsize=(16, 9))
        f.subplots_adjust(left=.08, right=.92, bottom=.05, top=.95, hspace=0)

        plt.show()

    return



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
feeps_split_integral_ch.py

@author : Louis RICHARD
"""

import xarray as xr


def feeps_split_integral_ch(inp_dataset=None):
    """This function splits the last integral channel from the FEEPS spectra, creating 2 new DataArrays

    Parameters
    ----------
    inp_dataset : xarray.DataArray
        Energetic particles energy spectrum from FEEPS.

    Returns
    -------
    out : xarray.DataArray
        Energetic particles energy spectra with the integral channel removed.

    out_500kev : xarray.DataArray
        Integral channel that was removed.

    """

    assert inp_dataset is not None and isinstance(inp_dataset, xr.Dataset)

    outdict, outdict_500kev = [{}, {}]

    for k in inp_dataset:
        try:
            # Energy spectra with the integral channel removed
            outdict[k] = inp_dataset[k][:, :-1]

            # Integral channel that was removed
            outdict_500kev[k] = inp_dataset[k][:, -1]
        except IndexError:
            pass

    out = xr.Dataset(outdict, attrs=inp_dataset.attrs)

    out_500kev = xr.Dataset(outdict_500kev, attrs=inp_dataset.attrs)

    return out, out_500kev



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
c_4_k.py

@author : Louis RICHARD
"""

import numpy as np
import xarray as xr

from .cross import cross
from .dot import dot


def c_4_k(r_list=None):
    """Calculates reciprocal vectors in barycentric coordinates.

    Parameters
    ----------
    r_list : list of xarray.DataArray
        Position of the spacecrafts.

    Returns
    -------
    k_list : list of xarray.DataArray
        Reciprocal vectors in barycentric coordinates.

    Notes
    -----
    The units of reciprocal vectors are the same as [1/r].

    """

    assert r_list is not None and isinstance(r_list, list) and len(r_list) == 4

    for i in range(4):
        if not isinstance(r_list[i], xr.DataArray):
            raise TypeError("Spacecraft position must be DataArray")

    mms_list = np.arange(4)

    k_list = [None] * 4

    for i, j, k, l in zip(mms_list, np.roll(mms_list, 1), np.roll(mms_list, 2),
                          np.roll(mms_list, 3)):
        cc = cross(r_list[k]-r_list[j], r_list[l]-r_list[j])

        dr12 = r_list[i]-r_list[j]

        k_list[i] = cc/dot(cc, dr12)

    return k_list



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
trace.py

@author : Louis RICHARD
"""

import xarray as xr


def trace(inp=None):
    """Computes trace of the time series of 2nd order tensors.

    Parameters
    ----------
    inp : xarray.DataArray
        Time series of the input 2nd order tensor.

    Returns
    -------
    out : xarray.DataArray
        Time series of the trace of the input tensor.

    Examples
    --------
    >>> from pyrfu import mms, pyrf
    >>> # Time interval
    >>> tint = ["2015-10-30T05:15:20.000", "2015-10-30T05:16:20.000"]
    >>> # Spacecraft index
    >>> mms_id = 1
    >>> # Load magnetic field and ion temperature
    >>> b_xyz = mms.get_data("B_gse_fgm_srvy_l2", tint, mms_id)
    >>> t_xyz_i = mms.get_data("Ti_gse_fpi_fast_l2", tint, mms_id)
    >>> # Rotate to ion temperature tensor to field aligned coordinates
    >>> t_xyzfac_i = mms.rotate_tensor(t_xyz_i, "fac", b_xyz, "pp")
    >>> # Compute scalar temperature
    >>> t_i = pyrf.trace(t_xyzfac_i)

    """

    assert inp is not None and isinstance(inp, xr.DataArray) and inp.data.ndim == 3

    inp_data = inp.data
    out_data = inp_data[:, 0, 0] + inp_data[:, 1, 1] + inp_data[:, 2, 2]

    # Attributes
    attrs = inp.attrs

    # Change tensor order from 2 (matrix) to 0 (scalar)
    attrs["TENSOR_ORDER"] = 0

    out = xr.DataArray(out_data, coords=[inp.time.data], dims=["time"], attrs=attrs)

    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
get_feeps_oneeye.py

@author : Louis RICHARD
"""

import numpy as np

from .get_feeps_active_eyes import get_feeps_active_eyes
from .db_get_ts import db_get_ts


def get_feeps_oneeye(tar_var="fluxe_brst_l2", e_id="bottom-4", tint=None, mms_id=1, verbose=True):
    """
    Load energy spectrum all the target eye

    Parameters
    ----------
    tar_var : str
        target variable "{data_units}{specie}_{data_rate}_{level}" :
            * data_units :
                * flux : intensity (1/cm sr).
                * count : counts (-).
                * CPS : counts per second (1/s).

            * specie :
                * i : ion.
                * e : electron.

            * data_rate :
                * brst : high resolution data.
                * srvy : low resolution data.

            * level :
                * l1 : level 1 data
                * l1b : level 1b data
                * l2 : level 2 data
                * l3 : level 3 data

    e_id : str
        index of the eye "{deck}-{id}" :
            * deck : top/bottom
            * id : see get_feeps_active_eyes

    tint : list of str
        Time interval.

    mms_id : int or str
        Index of the spacecraft.

    verbose : bool, optional
        Set to True to follow the loading. Default is True.

    Returns
    -------
    out : xarray.DataArray
        Energy spectrum of the target eye.

    """

    assert isinstance(tar_var, str)
    assert isinstance(e_id, str)
    assert tint is not None and isinstance(tint, list)
    assert isinstance(tint[0], str) and isinstance(tint[1], str)
    assert isinstance(mms_id, (int, str)) and int(mms_id) in np.arange(1, 5)
    assert isinstance(verbose, bool)

    mms_id = int(mms_id)

    var = {"inst": "feeps"}

    data_units = tar_var.split("_")[0][:-1]
    specie = tar_var.split("_")[0][-1]

    if specie == "e":
        var["dtype"] = "electron"
    elif specie == "i":
        var["dtype"] = "ion"
    else:
        raise ValueError("invalid specie")

    var["tmmode"] = tar_var.split("_")[1]
    var["lev"] = tar_var.split("_")[2]

    dset_name = f"mms{mms_id:d}_feeps_{var['tmmode']}_l2_{var['dtype']}"
    dset_pref = f"mms{mms_id:d}_epd_feeps_{var['tmmode']}_{var['lev']}_{var['dtype']}"

    active_eyes = get_feeps_active_eyes(var, tint, mms_id)

    if e_id.split("-")[0] in ["top", "bottom"]:
        suf = e_id.split("-")[0]

        e_id = int(e_id.split("-")[1])

        if e_id in active_eyes[suf]:
            if data_units.lower() == "flux":
                suf = "_".join([suf, "intensity", "sensorid", str(e_id)])
            elif data_units.lower() == "counts":
                suf = "_".join([suf, "counts", "sensorid", str(e_id)])
            elif data_units.lower() == "cps":
                suf = "_".join([suf, "count_rate", "sensorid", str(e_id)])
            elif data_units == "mask":
                suf = "_".join([suf, "sector_mask", "sensorid", str(e_id)])
            else:
                raise ValueError("undefined variable")
        else:
            raise ValueError("Unactive eye")
    else:
        raise ValueError("Invalid format of eye id")

    if verbose:
        print("Loading {}...".format("_".join([dset_pref, suf])))

    out = db_get_ts(dset_name, "_".join([dset_pref, suf]), tint)

    out.attrs["tmmode"] = var["tmmode"]
    out.attrs["lev"] = var["lev"]
    out.attrs["mms_id"] = mms_id
    out.attrs["dtype"] = var["dtype"]
    out.attrs["species"] = "{}s".format(var["dtype"])
    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
deflux_to_vdf.py

@author : Louis RICHARD
"""

import numpy as np
import xarray as xr

from astropy import constants

from .spectr_to_dataset import spectr_to_dataset


def deflux_to_vdf(deflux=None):
    """Compute density energy flux from velocity distribution function.

    Parameters
    ----------
    deflux : xarray.Dataset or xarray.DataArray
        Time series of the density energy flux in 1/(cm^2 s sr keV).

    Returns
    -------
    vdf : xarray.Dataset
        Time series of the velocity distribution function. Units must be either s^3/m^6.

    """

    assert deflux is not None and isinstance(deflux, (xr.Dataset, xr.DataArray))

    if isinstance(deflux, xr.DataArray):
        deflux = spectr_to_dataset(deflux)

    if deflux.attrs["species"] in ["ions", "i"]:
        mm = 1
    elif deflux.attrs["species"] in ["electrons", "e"]:
        mm = constants.m_e.value / constants.m_p.value
    else:
        raise ValueError("Invalid specie")

    if deflux.attrs["UNITS"] == "keV/(cm^2 s sr keV)":
        tmp_data = deflux.data.data / 1e12 * 0.53707 * mm ** 2
    else:
        raise ValueError("Invalid unit")

    energy = deflux.energy.data

    if tmp_data.ndim == 2:
        tmp_data = tmp_data[:, :, None, None]

    data_r = np.reshape(tmp_data,
                        (tmp_data.shape[0], tmp_data.shape[1], np.prod(tmp_data.shape[2:])))

    if energy.ndim == 1:
        energy_mat = np.tile(energy, (len(deflux.time), np.prod(tmp_data.shape[2:]), 1))
        energy_mat = np.transpose(energy_mat, [0, 2, 1])
    elif energy.ndim == 2:
        energy_mat = np.tile(energy, (np.prod(tmp_data.shape[2:]), 1, 1))
        energy_mat = np.transpose(energy_mat, [1, 2, 0])
    else:
        raise ValueError("Invalid energy shape")

    data_r /= energy_mat ** 2
    tmp_data = np.reshape(data_r, tmp_data.shape)

    vdf = deflux.copy()
    vdf.data.data = np.squeeze(tmp_data)
    vdf.attrs["UNITS"] = "s^3/m^6"

    return vdf



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
avg_4sc.py
@author : Louis RICHARD
"""

import xarray as xr

from .resample import resample


def avg_4sc(b_list=None):
    """Computes the input quantity at the center of mass of the MMS tetrahedron.

    Parameters
    ----------
    b_list : list of xarray.DataArray
        List of the time series of the quantity for each spacecraft.

    Returns
    -------
    b_avg : xarray.DataArray
        Time series of the input quantity a the enter of mass of the MMS tetrahedron.

    Examples
    --------
    >>> from pyrfu.mms import get_data
    >>> from pyrfu.pyrf import avg_4sc
    >>> # Time interval
    >>> tint = ["2019-09-14T07:54:00.000", "2019-09-14T08:11:00.000"]
    >>> # Spacecraft indices
    >>> b_mms = [mms.get_data("B_gse_fgm_srvy_l2", tint, mms_id) for mms_id in range(1, 5)]
    >>> b_xyz = pyrf.avg_4sc(b_mms)

    """

    if b_list is None:
        raise ValueError("avg_4sc requires 1 argument")

    if not isinstance(b_list, list):
        raise TypeError("B must be a list of the 4 spacecraft data")

    for i, b in enumerate(b_list):
        if not isinstance(b, xr.DataArray):
            raise TypeError("B[{:d}] must be a DataArray".format(i))

    b_list = [resample(b, b_list[0]) for b in b_list]

    b_avg = sum(b_list)/len(b_list)

    return b_avg



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
start.py

@author : Louis RICHARD
"""

import xarray as xr
from astropy.time import Time


def start(inp=None):
    """Gives the first time of the time series.

    Parameters
    ----------
    inp : xarray.DataArray
        Time series.

    Returns
    -------
    out : float or str
        Value of the first time in the desired format.

    """

    assert inp is not None and isinstance(inp, xr.DataArray)

    out = Time(inp.time.data[0], format="datetime64").unix
    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
correct_edp_probe_timing.py

@author : Louis RICHARD
"""

import numpy as np
import xarray as xr

from ..pyrf import ts_scalar, resample


def correct_edp_probe_timing(v0=None):
    """Corrects for the channel delays not accounted for in the MMS EDP processing. As described
    in the MMS EDP data products guide.

    Parameters
    ----------
    v0 : xarray.DataArray
        Time series created from L2 sc_pot files, from the variable "mms#_edp_dcv_brst_l2"
        containing individual probe potentials.

    Returns
    -------
    v_corrected : xarray.DataArray
        Time series where the channel delay for each probe have been accounted and corrected for.

    Notes
    -----
    This function is only useful for Burst mode data. For the other telemetry modes (i.e. slow
    and fast) the channel delays are completely negligible and the interpolation and resampling
    applied here will have no effect other than possibly introduce numerical noise.

    """

    # Verify input
    assert v0 is not None and isinstance(v0, xr.DataArray)

    e_fact = [.1200, .1200, .0292]

    # Reconstruct E12, E34, E56 as computed in MMS processing
    t = v0.time.data
    es = [ts_scalar(t, np.diff(v0.data[:, i:i + 2]) / fact) for i, fact in zip([0, 2, 4], e_fact)]

    # Correct the time tags to create individual time series
    tau_vs = [np.timedelta64(0, "ns"), np.timedelta64(7629, "ns"), np.timedelta64(15259, "ns")]
    tau_es = [np.timedelta64(26703, "ns"), np.timedelta64(30518, "ns"), np.timedelta64(34332, "ns")]

    # Odds probes potential 1, 3, 5
    vo = [ts_scalar(t + tau, v0.data[:, i]) for tau, i in zip(tau_vs, [0, 2, 4])]

    # Electric field
    es = [ts_scalar(e.time.data + tau, e.data) for e, tau in zip(es, tau_es)]

    # Resample all data to time tags of V1 (i.e. timeOrig).
    vo = [resample(v, v0) for v in vo]
    es = [resample(e, v0) for e in es]

    # Recompute individual even probe potentials 2, 4, 6
    ve = [v - e * fact for v, e, fact in zip(vo, es, e_fact)]

    v = [item.data for tup in zip(ve, vo) for item in tup]

    # Create the new time series with the corrected values
    probe_index = np.arange(1, 7)
    v_corrected = xr.DataArray(v, coords=[t, probe_index], dims=["time", "probe"])

    return v_corrected



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
get_pitch_angle_dist.py

@author : Louis RICHARD
"""

import numpy as np
import xarray as xr

from pyrfu.pyrf import ts_scalar, time_clip, resample, normalize


def get_pitch_angle_dist(vdf=None, b_xyz=None, tint=None, **kwargs):
    """
    Computes the pitch angle distributions from l1b brst particle data.

    Parameters
    ----------
    vdf : xarray.Dataset
        to fill

    b_xyz : xarray.DataArray
        to fill

    tint : list of str, optional
        Time interval for closeup.

    **kwargs : dict
        Hash table of keyword arguments with :
            * angles : int or float or list of numpy.ndarray
                User defined angles

            * meanorsum : str
                Method :
                    * mean : to fill
                    * sum : to fill
                    * sum_weighted : to fill


    Returns
    -------
    pad : xarray.DataArray
        Particle pitch angle distribution


    Examples
    --------
    >>> from pyrfu import mms
    >>> # Define time interval
    >>> tint_long = ["2017-07-24T12:48:34.000", "2017-07-24T12:58:20.000"]
    >>> # Load ions velocity distribution for MMS1
    >>> vdf_i = mms.get_data("pdi_fpi_brst_l2", tint_long, 1)
    >>> # Load magnetic field in the spacecraft coordinates system.
    >>> b_dmpa = mms.get_data("b_dmpa_fgm_brst_l2", tint_long, 1)
    >>> # Define closeup time interval
    >>> tint_zoom = ["2017-07-24T12:49:18.000", "2017-07-24T12:49:30.000"]
    >>> # Compute pitch angle distribution
    >>> options = dict(anggles=24)
    >>> pad_i = mms.get_pitch_angle_dist(vdf, b_dmpa, tint_zoom, **options)

    """

    assert vdf is not None and isinstance(vdf, xr.Dataset)
    assert b_xyz is not None and isinstance(b_xyz, xr.Dataset)
    assert tint is not None and isinstance(tint, list)
    assert isinstance(tint[0], str) and isinstance(tint[1], str)

    # Default pitch angles. 15 degree angle widths
    angles_v = np.linspace(15, 180, int(180 / 15))
    d_angles = np.median(np.diff(angles_v)) * np.ones(len(angles_v))

    # Method
    mean_or_sum = "mean"

    if "angles" in kwargs:
        if isinstance(kwargs["angles"], (int, float)):
            n_angles = np.floor(kwargs["angles"])  # Make sure input is integer
            d_angles = 180 / n_angles
            angles_v = np.linspace(d_angles, 180, n_angles)
            d_angles = d_angles * np.ones(n_angles)
            print("notice : User defined number of pitch angles.")

        elif isinstance(kwargs["angles"], (list, np.ndarray)):
            angles_v = kwargs["angles"]
            d_angles = np.diff(angles_v)
            angles_v = angles_v[1:]
            print("notice : User defined pitch angle limits.")

        else:
            raise ValueError("angles parameter not understood.")

    if "meanorsum" in kwargs:
        if isinstance(kwargs["meanorsum"], str) and kwargs["meanorsum"] in ["mean", "sum"]:
            mean_or_sum = kwargs["meanorsum"]
        else:
            raise ValueError("meanorsum parameter not understood.")

    pitch_angles = angles_v - d_angles / 2

    time = vdf.time.data

    vdf0 = vdf.data

    if vdf.phi.data.ndim == 1:
        phi = np.tile(vdf.phi.data, (len(time), 1))
        phi = xr.DataArray(phi, coords=[time, np.arange(len(phi))], dims=["time", "idx1"])
    else:
        phi = vdf.phi

    theta = vdf.theta.data

    if "esteptable" in vdf.attrs.keys():
        step_table = ts_scalar(time, vdf.attrs["esteptable"])
    else:
        step_table = ts_scalar(time, np.zeros(len(time)))

    if "energy0" in vdf.attrs.keys() and "energy1" in vdf.attrs.keys():
        energy0, _ = [vdf.attrs[f"energy{i}"] for i in range(2)]
    else:
        energy0, _ = vdf.energy.data[:2, :]

    if tint is not None:
        b_xyz = time_clip(b_xyz, tint)
        vdf0, phi, step_table = [time_clip(dat, tint) for dat in [vdf0, phi, step_table]]

    time = vdf0.time.data

    # Check size of energy
    n_en, n_phi, n_theta = [len(energy0), len(phi.data[0, :]), len(theta)]

    b_xyz = resample(b_xyz, vdf0)
    b_vec = normalize(b_xyz)

    b_vec_x = np.transpose(np.tile(b_vec.data[:, 0], [n_en, n_phi, n_theta, 1]), [3, 0, 1, 2])
    b_vec_y = np.transpose(np.tile(b_vec.data[:, 1], [n_en, n_phi, n_theta, 1]), [3, 0, 1, 2])
    b_vec_z = np.transpose(np.tile(b_vec.data[:, 2], [n_en, n_phi, n_theta, 1]), [3, 0, 1, 2])

    x = np.zeros((len(time), n_phi, n_theta))
    y = np.zeros((len(time), n_phi, n_theta))
    z = np.zeros((len(time), n_phi, n_theta))

    for ii in range(len(time)):
        x[ii, ...] = np.dot(-np.cos(phi.data[ii, None] * np.pi / 180).T,
                            np.sin(theta.data[:, None] * np.pi / 180).T)
        y[ii, ...] = np.dot(-np.sin(phi.data[ii, None] * np.pi / 180).T,
                            np.sin(theta.data[:, None] * np.pi / 180).T)
        z[ii, ...] = np.dot(-np.ones((n_phi, 1)), np.cos(theta.data[:, None] * np.pi / 180).T)

    if tint is not None:
        energy = time_clip(vdf.energy, tint).data
    else:
        energy = vdf.energy.data

    xt, yt, zt = [np.tile(mat, [n_en, 1, 1, 1]) for mat in [x, y, z]]
    xt, yt, zt = [np.squeeze(np.transpose(mat, [1, 0, 2, 3])) for mat in [xt, yt, zt]]

    theta_b = np.arccos(xt * np.squeeze(b_vec_x) + yt * np.squeeze(b_vec_y) + zt * np.squeeze(
        b_vec_z)) * 180 / np.pi

    dists = [vdf0.data for _ in range(len(angles_v))]

    pad_arr = []

    for i, dist in enumerate(dists):
        dist[theta_b < angles_v[i] - d_angles[i]] = np.nan
        dist[theta_b > angles_v[i]] = np.nan
        if mean_or_sum == "mean":
            pad_arr.append(np.squeeze(np.nanmean(np.nanmean(dist, axis=3), axis=2)))
        elif mean_or_sum == "sum":
            pad_arr.append(np.squeeze(np.nansum(np.nansum(dist, axis=3), axis=2)))
        else:
            raise ValueError("Invalid method")

    pad_arr = np.stack(pad_arr)
    pad_arr = np.transpose(pad_arr, [1, 0, 2])

    energy = np.mean(energy[:2, :], axis=0)

    pad = xr.DataArray(pad_arr, coords=[time, pitch_angles, energy],
                       dims=["time", "theta", "energy"])

    pad.attrs = vdf.attrs
    pad.attrs["mean_or_sum"] = mean_or_sum
    pad.attrs["delta_pitchangle_minus"] = d_angles * .5
    pad.attrs["delta_pitchangle_plus"] = d_angles * .5

    return pad



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
dot.py

@author : Louis RICHARD
"""

import numpy as np
import xarray as xr

from .resample import resample
from .ts_scalar import ts_scalar


def dot(x=None, y=None):
    """Computes dot product of two fields.

    Parameters
    ----------
    x : xarray.DataArray
        Time series of the first field X.

    y : xarray.DataArray
        Time series of the second field Y.

    Returns
    -------
    out : xarray.DataArray
        Time series of the dot product Z = X.Y.

    Examples
    --------
    >>> from pyrfu import mms, pyrf
    >>> # Time interval
    >>> tint = ["2019-09-14T07:54:00.000", "2019-09-14T08:11:00.000"]
    >>> # Load magnetic field, electric field and spacecraft position
    >>> r_mms, b_mms, e_mms = [[] * 4 for _ in range(3)]
    >>> for mms_id in range(1, 5):
    >>>		r_mms.append(mms.get_data("R_gse", tint, mms_id))
    >>> 	b_mms.append(mms.get_data("B_gse_fgm_srvy_l2", tint, mms_id))
    >>>		e_mms.append(mms.get_data("E_gse_edp_fast_l2", tint, mms_id))
    >>>
    >>> j_xyz, div_b, b_avg, jxb, div_t_shear, div_pb = pyrf.c_4_j(r_mms, b_mms)
    >>> # Compute the electric at the center of mass of the tetrahedron
    >>> e_xyz = pyrf.avg_4sc(e_mms)
    >>> # Compute J.E dissipation
    >>> je = pyrf.dot(j_xyz, e_xyz)

    """

    assert x is not None and isinstance(x, xr.DataArray)
    assert y is not None and isinstance(y, xr.DataArray)

    # Resample to first input sampling
    y = resample(y, x)

    # Compute scalar product
    out_data = np.sum(x.data * y.data, axis=1)

    # Output to xarray
    out = ts_scalar(x.time.data, out_data)

    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
vht.py

@author : Louis RICHARD
"""

import numpy as np
import xarray as xr

from .resample import resample
from .e_vxb import e_vxb


def vht(e=None, b=None, flag=1):
    """
    Estimate velocity of the De Hoffman-Teller frame from the velocity estimate the electric field
    eht=-vht x b

    Parameters
    ----------
    e : xarray.DataArray
        Time series of the electric field.

    b : xarray.DataArray
        Time series of the magnetic field.

    flag : int
        If 2 assumed no Ez.

    Returns
    -------
    vht : numpy.ndarray
        De Hoffman Teller frame velocity [km/s].

    vht : xarray.DataArray
        Time series of the electric field in the De Hoffman frame.

    dv_ht : numpy.ndarray
        Error of De Hoffman Teller frame.

    """

    assert e is not None and isinstance(e, xr.DataArray)
    assert b is not None and isinstance(b, xr.DataArray)

    n_samples = len(e)

    # Resample magnetic field to electric field sampling (usually higher)

    if n_samples != len(b):
        b = resample(b, e)

    p = np.zeros(6)

    p[0] = np.sum(b[:, 0].data * b[:, 0].data) / n_samples  # Bx*Bx
    p[1] = np.sum(b[:, 0].data * b[:, 1].data) / n_samples  # Bx*By
    p[2] = np.sum(b[:, 0].data * b[:, 2].data) / n_samples  # Bx*Bz
    p[3] = np.sum(b[:, 1].data * b[:, 1].data) / n_samples  # By*By
    p[4] = np.sum(b[:, 1].data * b[:, 2].data) / n_samples  # By*Bz
    p[5] = np.sum(b[:, 2].data * b[:, 2].data) / n_samples  # Bz*Bz

    # assume only Ex and Ey
    if flag == 2:
        e[:, 2] *= 0     # put z component to 0 when using only Ex and Ey

        k_mat = np.array([[p[5], 0, -p[2]], [0, p[5], -p[4]], [-p[2], -p[4], p[0] + p[3]]])
        u_msg = "De Hoffman-Teller frame is calculated using 2 components of E=(Ex,Ey,0)"
    else:
        k_mat = np.array(
            [[p[3] + p[5], -p[1], -p[2]], [-p[1], p[0] + p[5], -p[4]], [-p[2], -p[4], p[0] + p[3]]])
        u_msg = "De Hoffman-Teller frame is calculated using all 3 components of E=(Ex,Ey,Ez)"

    exb = np.cross(e, b)

    ind_data = np.where(~np.isnan(exb[:, 0].data))[0]

    # revised by Wenya LI; 2015-11-21, wyli @ irfu
    tmp1 = exb[ind_data]
    exb_avg = np.sum(tmp1, axis=0) / n_samples
    # averExB=sum(ExB(indData).data,1)/nSamples;
    # end revise.
    v_ht = np.linalg.solve(k_mat, exb_avg.T) * 1e3  # 9.12 in ISSI book

    v_ht_hat = v_ht/np.linalg.norm(v_ht, keepdims=True)

    print(u_msg)
    print("v_ht ={:7.4f} * {} = {} km/s".format(np.linalg.norm(v_ht), np.array_str(v_ht_hat),
                                                np.array_str(v_ht)))

    # Calculate the goodness of the Hoffman Teller frame
    e_ht = e_vxb(v_ht, b)

    if flag == 2:
        e_p, e_ht_p = [e[ind_data], e_ht[ind_data]]
        e_p.data[:, 2], e_ht_p.data[:, 2] = [0, 0]
    else:
        e_p, e_ht_p = [e[ind_data], e_ht[ind_data]]

    delta_e = e_p.data - e_ht_p.data

    poly_fit = np.polyfit(e_ht_p.data.reshape([len(e_ht_p) * 3]), e_p.data.reshape([len(e_p) * 3]),
                          1)
    corr_coeff = np.corrcoef(e_ht_p.data.reshape([len(e_ht_p) * 3]),
                             e_p.data.reshape([len(e_p) * 3]))

    print("slope = {p[0]:6.4f}, offs = {p[1]:6.4f}".format(p=poly_fit))
    print("cc = {:6.4f}".format(corr_coeff[0, 1]))

    dv_ht = np.sum(np.sum(delta_e ** 2)) / len(ind_data)
    s_mat = (dv_ht / (2 * len(ind_data) - 3)) / k_mat
    dv_ht = np.sqrt(np.diag(s_mat)) * 1e3

    dv_ht_hat = dv_ht / np.linalg.norm(dv_ht)

    print("dv_ht ={:7.4f} * {} = {} km/s".format(np.linalg.norm(dv_ht), np.array_str(dv_ht_hat),
                                                 np.array_str(dv_ht)))

    return v_ht, e_ht, dv_ht



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
histogram2d.py

@author : Louis RICHARD
"""

import numpy as np
import xarray as xr

from .resample import resample


def histogram2d(inp1=None, inp2=None, bins=100):
    """Computes 2d histogram of inp2 vs inp1 with nbins number of bins.

    Parameters
    ----------
    inp1 : xarray.DataArray
        Time series of the x values.

    inp2 : xarray.DataArray
        Time series of the y values.

    bins : int
        Number of bins.

    Returns
    -------
    out : xarray.DataArray
        2D map of the density of inp2 vs inp1.

    Examples
    --------
    >>> from pyrfu import mms, pyrf
    >>> # Time interval
    >>> tint = ["2019-09-14T07:54:00.000", "2019-09-14T08:11:00.000"]
    >>> # Spacecraft indices
    >>> mms_id = 1
    >>> # Load magnetic field and electric field
    >>> b = mms.get_data("B_gse_fgm_srvy_l2", tint, mms_id)
    >>> r = mms.get_data("R_gse", tint, mms_id)
    >>> # Compute current density, etc
    >>> j_xyz, div_b, b_xyz, jxb, div_t_shear, div_pb = pyrf.c_4_j(r, b)
    >>> # Compute magnitude of B and J
    >>> b_mag = pyrf.norm(b_xyz)
    >>> j_mag = pyrf.norm(j_xyz)
    >>> # Histogram of |J| vs |B|
    >>> h2d_b_j = pyrf.histogram2d(b_mag, j_mag)

    """

    assert inp1 is not None and isinstance(inp1, xr.DataArray)
    assert inp2 is not None and isinstance(inp2, xr.DataArray)

    # resample inp2 with respect to inp1
    if len(inp2) != len(inp1):
        inp2 = resample(inp2, inp1)

    h2d, x_edges, y_edges = np.histogram2d(inp1.data, inp2.data, bins=bins)

    x = x_edges[:-1] + np.median(np.diff(x_edges)) / 2
    y = y_edges[:-1] + np.median(np.diff(y_edges)) / 2

    out = xr.DataArray(h2d, coords=[x, y], dims=["x_bins", "y_bins"])

    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
zoom.py

@author : Louis RICHARD
"""

from matplotlib.transforms import Bbox, TransformedBbox, blended_transform_factory
from mpl_toolkits.axes_grid1.inset_locator import BboxPatch, BboxConnector, BboxConnectorPatch

from pandas.plotting import register_matplotlib_converters
register_matplotlib_converters()


def connect_bbox(bbox1, bbox2, loc1a, loc2a, loc1b, loc2b, prop_lines, prop_patches=None):
    if prop_patches is None:
        prop_patches = {**prop_lines, "alpha": prop_lines.get("alpha", 1) * 0}

    c1 = BboxConnector(bbox1, bbox2, loc1=loc1a, loc2=loc2a, **prop_lines)
    c1.set_clip_on(False)
    c2 = BboxConnector(bbox1, bbox2, loc1=loc1b, loc2=loc2b, **prop_lines)
    c2.set_clip_on(False)

    bbox_patch1 = BboxPatch(bbox1, **prop_patches)
    bbox_patch2 = BboxPatch(bbox2, **prop_patches)

    p = BboxConnectorPatch(bbox1, bbox2, loc1a=loc1a, loc2a=loc2a, loc1b=loc1b, loc2b=loc2b,
                           **prop_patches)

    p.set_clip_on(False)

    return c1, c2, bbox_patch1, bbox_patch2, p


def zoom(ax1, ax2, **kwargs):
    """Similar to zoom_effect01.  The xmin & xmax will be taken from the ax1.viewLim.

    Parameters
    ----------
    ax1 : axs
        The main axes.

    ax2 : axs
        The zoomed axes.

    """

    tt = ax1.transScale + (ax1.transLimits + ax2.transAxes)
    trans = blended_transform_factory(ax2.transData, tt)

    mybbox1 = ax1.bbox
    mybbox2 = TransformedBbox(ax1.viewLim, trans)

    prop_patches = {**kwargs, "ec": "none", "alpha": 0.2}

    c1, c2, bbox_patch1, bbox_patch2, p = connect_bbox(mybbox1, mybbox2, loc1a=2, loc2a=3,
                                                       loc1b=1, loc2b=4, prop_lines=kwargs)

    ax1.add_patch(bbox_patch1)
    ax2.add_patch(bbox_patch2)
    ax2.add_patch(c1)
    ax2.add_patch(c2)
    ax2.add_patch(p)

    return



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
dispersion_surface_calc.py

@author : Louis RICHARD
"""

import numpy as np


def dispersion_surface_calc(kc_x_max=None, kc_z_max=None, m_i=None, w_pe=None):
    """Calculates the cold plasma dispersion surfaces according to equation 2.63 in [15]_, and puts
    them in the matrix W. Additional parameters that are needed in dispersion_surface is returned
    as extra_param.

    Parameters
    ----------
    kc_x_max : float or int
        Max value of k_perpendicular*c/w_c.

    kc_z_max : float or int
        Max value of k_parallel*c/w_c.

    m_i : float
        Ion mass in terms of electron masses.

    w_pe : float
        Electron plasma frequency in terms of electron gyro frequency.

    Returns
    -------
    w_final : numpy.ndarray
        to fill.

    extra_param : numpy.ndarray
        to fill.

    Notes
    -----
    This function is essential for dispersion_surface to work.

    References
    ----------
    .. [15] Swanson, D. G. (2003), Plasma waves, 2nd ed., Institute of Physics Pub,
            doi: https://www.taylorfrancis.com/books/9780367802721

    """

    assert kc_x_max is not None and isinstance(kc_x_max, (float, int))
    assert kc_z_max is not None and isinstance(kc_z_max, (float, int))
    
    if isinstance(kc_x_max, float):
        kc_x_max = int(kc_x_max)
    
    if isinstance(kc_z_max, float):
        kc_z_max = int(kc_z_max)
        
    # First get rid of those annoying "division by zero" - warnings

    # Make vectors of the wave numbers
    kc_x, kc_z = [np.linspace(0.000001, kc_max, 35) for kc_max in [kc_x_max, kc_z_max]]

    # Turn those vectors into matrices
    [kc_x_mat, kc_z_mat] = np.meshgrid(kc_x, kc_z)

    # Find some of the numbers that appear later in the calculations
    kc = np.sqrt(kc_x_mat ** 2 + kc_z_mat ** 2)     # Absolute value of k
    theta = np.arctan2(kc_x_mat, kc_z_mat)          # The angle between k and B

    w_ci = 1 / m_i                                  # The ion gyro frequency
    w_pi = w_pe / np.sqrt(m_i)                      # The ion plasma frequency
    w_p = np.sqrt(w_pe ** 2 + w_pi ** 2)            # The total plasma frequency

    # To speed up the program somewhat introduce these
    kc2 = kc ** 2
    kc4 = kc2 ** 2
    w_ci2 = w_ci ** 2
    w_p2 = w_p ** 2
    w_extra = w_p2 + w_ci
    w_extra2 = w_extra ** 2
    cos_theta2 = np.cos(theta) ** 2

    # For every k_perp and k_par, turn the dispersion relation into a polynomial eq. and solve it.
    # The polynomial coefficients are calculated
    pol_k_coeff8 = -(2 * kc2 + 1 + w_ci2 + 3 * w_p2)
    pol_k_coeff6 = (kc4 + (2 * kc2 + w_p2) * (1 + w_ci2 + 2 * w_p2) + w_extra2)
    pol_k_coeff4 = -(
                kc4 * (1 + w_ci2 + w_p2) + 2 * kc2 * w_extra2 + kc2 * w_p2 * (1 + w_ci2 - w_ci) * (
                    1 + cos_theta2) + w_p2 * w_extra2)
    pol_k_coeff2 = (kc4 * (w_p2 * (
                1 + w_ci2 - w_ci) * cos_theta2 + w_ci * w_extra) + kc2 * w_p2 * w_ci * w_extra * (
                                1 + cos_theta2))
    pol_k_coeff0 = -kc4 * w_ci2 * w_p2 * cos_theta2

    # For each k, solve the equation
    w_final = np.zeros((10, 35, 35))
    for k_z in range(len(kc_z)):
        for k_x in range(len(kc_x)):
            dispersion_polynomial = [1, 0, pol_k_coeff8[k_z, k_x], 0, pol_k_coeff6[k_z, k_x], 0,
                                     pol_k_coeff4[k_z, k_x], 0, pol_k_coeff2[k_z, k_x], 0,
                                     pol_k_coeff0[k_z, k_x]]
            # theoretically should be real (A. Tjulin)
            w_temp = np.real(np.roots(dispersion_polynomial))
            # We need to sort the answers to get nice surfaces.
            w_final[:, k_z, k_x] = np.sort(w_temp)

    # Now we have solved the dispersion relation. Let us find some other interesting parameters in
    # this context.
    # The elements of the dielectric tensor, using Swanson's notation
    dielectric_s = 1 - w_pe ** 2 / (w_final ** 2 - 1) - w_pi ** 2 / (w_final ** 2 - w_ci2)
    dielectric_p = 1 - (w_pe ** 2 + w_pi ** 2) / (w_final ** 2)
    dielectric_d = -w_pe ** 2 / (w_final * (w_final ** 2 - 1)) + w_ci * w_pi ** 2 / (
                w_final * (w_final ** 2 - w_ci2))

    # The rest of this function is not cleaned yet. The calculations could probably be much
    # shorter ! Use tile instead.
    kc2_mat, k_x_mat, k_z_mat, theta_mat = [np.zeros((10, 35, 35)) for _ in range(4)]

    for i in range(10):
        kc2_mat[i, :, :] = kc2
        k_x_mat[i, :, :] = kc_x_mat
        k_z_mat[i, :, :] = kc_z_mat
        theta_mat[i, :, :] = theta

    n2 = kc2_mat / (w_final ** 2)
    v_phase_to_c = np.sqrt(1 / n2)
    va_to_c = 1 / (w_pe * np.sqrt(m_i))
    v_phase_to_va = v_phase_to_c / va_to_c

    # dielectric_xx = dielectric_s - n2 * np.cos(theta_mat) ** 2
    dielectric_xy = -1j * dielectric_d
    dielectric_xz = n2 * np.cos(theta_mat) * np.sin(theta_mat)
    dielectric_yy = dielectric_s - n2
    dielectric_zz = dielectric_p - n2 * np.sin(theta_mat) ** 2

    e_x = - dielectric_zz / dielectric_xz
    e_y = dielectric_xy / dielectric_yy * e_x
    e_z = 1
    
    e_perp = np.sqrt(e_x * np.conj(e_x) + e_y * np.conj(e_y))
    e_tot = np.sqrt(e_x * np.conj(e_x) + e_y * np.conj(e_y) + 1)
    e_par_k = (k_x_mat * e_x + k_z_mat * e_z) / np.sqrt(kc2_mat)
    e_polar = -2 * np.imag(e_x * np.conj(e_y)) / e_perp ** 2

    b_x = -k_z_mat * e_y / w_final
    b_y = (k_z_mat * e_x - k_x_mat * e_z) / w_final
    b_z = k_x_mat * e_y / w_final
    b_tot = np.sqrt(b_x * np.conj(b_x) + b_y * np.conj(b_y) + b_z * np.conj(b_z))
    b_par = np.sqrt(b_z * np.conj(b_z))
    b_perp = np.sqrt(b_x * np.conj(b_x) + b_y * np.conj(b_y))
    b_polar = -2 * np.imag(b_x * np.conj(b_y)) / (b_perp ** 2)

    # Poynting flux
    s_x = e_y * np.conj(b_z) - e_z * np.conj(b_y)
    s_y = e_z * np.conj(b_x) - e_x * np.conj(b_z)
    s_z = e_x * np.conj(b_y) - e_y * np.conj(b_x)
    s_par = abs(s_z)
    s_tot = np.sqrt(s_x * np.conj(s_x) + s_y * np.conj(s_y) + s_z * np.conj(s_z))

    temp = len(kc_x)
    dk_x = kc_x[1]
    dk_z = kc_z[1]
    dw_x = np.diff(w_final, 1, 3)
    dw_z = np.diff(w_final, 1, 2)
    dw_x[1, temp, temp] = 0
    dw_z[1, temp, temp] = 0
    v_x = dw_x / dk_x
    v_z = dw_z / dk_z

    # Compute ion and electron velocities
    q_e, q_i, m_e, w_ce = [-1, 1, 1, 1]

    v_x_e = 1j * q_e / (m_e * (w_final ** 2 - w_ce ** 2)) * (w_final * e_x - 1j * w_ce * e_y)
    v_y_e = 1j * q_e / (m_e * (w_final ** 2 - w_ce ** 2)) * (w_final * e_y + 1j * w_ce * e_x)
    v_z_e = 1j * q_e * e_z / (m_e * w_final)

    v_x_i = 1j * q_i / (m_i * (w_final ** 2 - w_ci ** 2)) * (w_final * e_x + 1j * w_ci * e_y)
    v_y_i = 1j * q_i / (m_i * (w_final ** 2 - w_ci ** 2)) * (w_final * e_y - 1j * w_ci * e_x)
    v_z_i = 1j * q_i * e_z / (m_i * w_final)

    # Ratio of parallel and perpendicular to B speed
    v_par_o_perp_e = v_z_e * np.conj(v_z_e) / (v_x_e * np.conj(v_x_e) + v_y_e * np.conj(v_y_e))
    v_par_o_perp_i = v_z_i * np.conj(v_z_i) / (v_x_i * np.conj(v_x_i) + v_y_i * np.conj(v_y_i))

    # Total particle speeds
    v_e2 = v_x_e * np.conj(v_x_e) + v_y_e * np.conj(v_y_e) + v_z_e * np.conj(v_z_e)
    v_i2 = v_x_i * np.conj(v_x_i) + v_y_i * np.conj(v_y_i) + v_z_i * np.conj(v_z_i)

    # Ion and electron energies
    e_e = 0.5 * m_e * v_e2
    e_i = 0.5 * m_i * v_i2

    # Ratio of particle and field energy densities
    ne = w_pe ** 2
    e_en = e_e * ne
    e_in = e_i * ne
    ee = 0.5 * e_tot ** 2
    eb = 0.5 * b_tot ** 2
    ratio_pf = (e_en + e_in) / (ee + eb)

    # Continuity equation
    dn_eon = (k_x_mat * v_x_e + k_z_mat * v_z_e) / w_final
    dn_eon = np.sqrt(dn_eon * np.conj(dn_eon))
    dn_ion = (k_x_mat * v_x_i + k_z_mat * v_z_i) / w_final
    dn_ion = np.sqrt(dn_ion * np.conj(dn_ion))
    dn_e_dn_i = dn_eon / dn_ion
    
    dn_eon_o_db_o_b = dn_eon / b_tot
    dn_ion_o_db_o_b = dn_ion / b_tot
    
    dn_eon_o_db_par_o_b = dn_eon / b_par
    dn_ion_o_db_par_o_b = dn_ion / b_par

    dn_e = dn_eon * w_pe ** 2
    k_dot_e = e_x * k_x_mat + e_z * k_z_mat
    k_dot_e = np.sqrt(k_dot_e * np.conj(k_dot_e))

    # Degree of electromagnetism
    extra_param = np.zeros((20, 10, 35, 35))
    extra_param[1, ...] = np.log10(b_tot / e_tot)

    # Degree of longitudinally
    extra_param[2, ...] = abs(e_par_k) / e_tot

    # Degree of parallelism
    extra_param[3, ...] = e_z / e_tot

    # Degree of parallelism
    extra_param[4, ...] = np.sqrt(b_z * np.conj(b_z)) / b_tot

    # Value of the group vel.
    extra_param[5, ...] = np.sqrt(v_x ** 2 + v_z ** 2)

    # Ellipticity
    extra_param[6, ...] = e_polar

    # Degree of electromagnetism
    extra_param[7, ...] = np.log10(e_tot ** 2 / b_tot ** 2)

    # Ratio of electron to ion energy
    extra_param[8, ...] = np.log10(ee / e_i)

    # Ratio of electron to ion velocity fluctuations
    extra_param[9, ...] = np.log10(np.sqrt(v_e2 / v_i2))

    # Ratio of particle to field energy densities
    extra_param[10, ...] = np.log10(ratio_pf)

    # Ellipticity based on B
    extra_param[11, ...] = b_polar

    # Phase speed divided by Alfven speed
    extra_param[12, ...] = np.log10(v_phase_to_va)

    # Ratio of parallel to perpendicular electron speed
    extra_param[13, ...] = np.log10(v_par_o_perp_e)

    # Ratio of parallel to perpendicular ion speed
    extra_param[14, ...] = np.log10(v_par_o_perp_i)

    #
    extra_param[15, ...] = np.log10(e_en / (ee + eb))

    # Ratio of electron to ion density perturbations
    extra_param[16, ...] = np.log10(dn_e_dn_i)

    # (dn_e / n) / (dB / B)
    extra_param[17, ...] = np.log10(dn_eon_o_db_o_b)

    # (dn_i / n) / (dB / B)
    extra_param[18, ...] = np.log10(dn_ion_o_db_o_b)

    # (dn_e / n) / (dB_par / B)
    extra_param[19, ...] = np.log10(dn_eon_o_db_par_o_b)

    # (dn_i / n) / (dB_par / B)
    extra_param[20, ...] = np.log10(dn_ion_o_db_par_o_b)

    # (dn_i / n) / (dB / B)
    extra_param[21, ...] = np.log10(dn_e / k_dot_e)

    # S_par / S_tot
    extra_param[22, ...] = s_par / s_tot

    return w_final, extra_param



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
t_eval.py

@author : Louis RICHARD
"""

import bisect
import numpy as np
import xarray as xr


def t_eval(inp=None, t=None):
    """Evaluates the input time series at the target time.

    Parameters
    ----------
    inp : xarray.DataArray
        Time series if the input to evaluate.

    t : numpy.ndarray
        Times at which the input will be evaluated.

    Returns
    -------
    out : DataArray
        Time series of the input at times t.

    """

    assert inp is not None and isinstance(inp, xr.DataArray)
    assert t is not None and isinstance(t, np.ndarray)

    idx = np.zeros(len(t))

    for i, time in enumerate(t):
        idx[i] = bisect.bisect_left(inp.time.data, time)

    idx = idx.astype(int)

    if inp.ndim == 2:
        out = xr.DataArray(inp.data[idx, :], coords=[t, inp.comp], dims=["time", "comp"])
    else:
        out = xr.DataArray(inp.data[idx], coords=[t], dims=["time"])

    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
spectr_to_dataset.py

@author : Louis RICHARD
"""

import numpy as np
import xarray as xr


def spectr_to_dataset(spectr=None):
    """Convert energy spectrum (DataArray) to Dataset.

    Parameters
    ----------
    spectr : xarray.DataArray
        Spectrogram in DataArray format.

    Returns
    -------
    out : xarray.Dataset
        Spectrogram in Dataset format.

    """

    assert spectr is not None and isinstance(spectr, xr.DataArray)

    time = spectr.time.data
    energy = spectr.energy.data
    energy = np.tile(energy, (len(time), 1))

    data = spectr.data

    out_dict = {"data": (["time", "idx0"], data), "energy": (["time", "idx0"], energy),
                "time": time, "idx0": np.arange(energy.shape[1])}

    out = xr.Dataset(out_dict)

    out.attrs = spectr.attrs

    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
convert_fac.py

@author : Louis RICHARD
"""

import numpy as np
import xarray as xr

from .resample import resample
from .ts_vec_xyz import ts_vec_xyz
from .calc_fs import calc_fs


def convert_fac(inp=None, b_bgd=None, r=np.array([1, 0, 0])):
    """
    Transforms to a field-aligned coordinate (FAC) system defined as :
        * R_parallel_z aligned with the background magnetic field
        * R_perp_y defined by R_parallel cross the position vector of the spacecraft (nominally
            eastward at the equator)
        * R_perp_x defined by R_perp_y cross R_par

    If inp is one vector along r direction, out is inp[perp, para] projection.

    Parameters
    ----------
    inp : xarray.DataArray
        Time series of the input field.

    b_bgd : xarray.DataArray
        Time series of the background magnetic field.

    r : xarray.DataArray or numpy.ndarray or list
        Position vector of spacecraft.

    Returns
    -------
    out : xarray.DataArray
        Time series of the input field in field aligned coordinates system.

    Notes
    -----
    All input parameters must be in the same coordinate system.

    Examples
    --------
    >>> import numpy
    >>> from pyrfu import mms, pyrf
    >>> # Time interval
    >>> tint = ["2019-09-14T07:54:00.000", "2019-09-14T08:11:00.000"]
    >>> # Spacecraft index
    >>> mms_id = 1
    >>> # Load magnetic field (FGM) and electric field (EDP)
    >>> b_xyz = mms.get_data("B_gse_fgm_brst_l2", tint, mms_id)
    >>> e_xyz = mms.get_data("E_gse_edp_brst_l2", tint, mms_id)
    >>> # Convert to field aligned coordinates
    >>> e_xyzfac = pyrf.convert_fac(e_xyz, b_xyz, numpy.array([1, 0, 0]))

    """

    assert inp is not None and isinstance(inp, xr.DataArray)
    assert b_bgd is not None and isinstance(b_bgd, xr.DataArray)

    if len(inp) != len(b_bgd):
        b_bgd = resample(b_bgd, inp, fs=calc_fs(inp))

    t, inp_data = [inp.time.data, inp.data]

    # Normalize background magnetic field
    bn = b_bgd / np.linalg.norm(b_bgd, axis=1, keepdims=True)

    if isinstance(r, (list, np.ndarray)) and len(r) == 3:
        r = np.tile(r, (len(b_bgd), 1))
    elif isinstance(r, xr.DataArray):
        r = resample(r, b_bgd)
    else:
        raise TypeError("Invalid type of spacecraft position")

    # Parallel
    r_par = bn

    # Perpendicular
    r_perp_y = np.cross(r_par, r, axis=1)
    r_perp_y /= np.linalg.norm(r_perp_y, axis=1, keepdims=True)
    r_perp_x = np.cross(r_perp_y, b_bgd, axis=1)
    r_perp_x /= np.linalg.norm(r_perp_x, axis=1, keepdims=True)

    n_data, n_dim = inp_data.shape

    if n_dim == 3:
        out_data = np.zeros(inp.shape)

        out_data[:, 0] = np.sum(r_perp_x * inp_data, axis=1)
        out_data[:, 1] = np.sum(r_perp_y * inp_data, axis=1)
        out_data[:, 2] = np.sum(r_par * inp_data, axis=1)

        # xarray
        out = xr.DataArray(out_data, coords=[t, inp.comp], dims=["time", "comp"])

    elif n_dim == 1:
        out_data = np.zeros([3, n_data])

        out_data[:, 0] = inp[:, 0] * (
                    r_perp_x[:, 0] * r[:, 0] + r_perp_x[:, 1] * r[:, 1] + r_perp_x[:, 2] * r[:, 2])
        out_data[:, 1] = inp[:, 0] * (
                    r_par[:, 0] * r[:, 0] + r_par[:, 1] * r[:, 1] + r_par[:, 2] * r[:, 2])

        out = ts_vec_xyz(t, out_data, attrs=inp.attrs)
    else:
        raise TypeError("Invalid dimension of inp")

    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
pl_scatter_matrix.py

@author : Louis RICHARD
"""

import xarray as xr
import warnings
import matplotlib.pyplot as plt

from ..pyrf import histogram2d
from . import plot_spectr


def pl_scatter_matrix(inp1=None, inp2=None, pdf=False, cmap="jet"):
	"""Produces a scatter plot of each components of field inp1 with respect to every component
	of field inp2. If pdf is set to True, the scatter plot becomes a 2d histogram.

	Parameters
	----------
	inp1 : xarray.DataArray
		First time series (x-axis).

	inp2 : xarray.DataArray
		Second time series (y-axis).

	m : str, optional
		Marker type. Default is "+". Not used if pdf is True

	pdf : bool
		Flag to plot the 2d histogram. If False the figure is a scatter plot. If True the figure
		is a 2d histogram.

	cmap : str
		Colormap. Default : "jet"

	Returns
	-------
	fig : matplotlib.pyplot.figure
		to fill.

	axs : matplotlib.pyplot.axes
		to fill.

	caxs : matplotlib.pyplot.colorbar
		Only if pdf is True

	"""

	if inp1 is None:
		raise ValueError("pl_scatter_matrix requires at least one argument")
	
	if inp2 is None:
		inp2 = inp1
		warnings.warn("inp2 is empty assuming that inp2=inp1", UserWarning)

	if not isinstance(inp1, xr.DataArray) or not isinstance(inp2, xr.DataArray):
		raise TypeError("Inputs must be DataArrays")
	
	if not pdf:
		fig, axs = plt.subplots(3, 3, sharex="all", sharey="all", figsize=(16, 9))
		fig.subplots_adjust(left=0.1, right=0.9, bottom=0.1, top=0.9, hspace=0.05, wspace=0.05)

		for i in range(3):
			for j in range(3):
				axs[j, i].scatter(inp1[:, i].data, inp2[:, j].data, marker="+")

		return fig, axs
	else:
		fig, axs = plt.subplots(3, 3, sharex="all", sharey="all", figsize=(16, 9))
		fig.subplots_adjust(left=0.1, right=0.9, bottom=0.1, top=0.9, hspace=0.05, wspace=0.3)

		hist = [[None]*3]*3

		caxs = [[None]*3]*3

		for i in range(3):
			for j in range(3):
				hist[j][i] = histogram2d(inp1[:, i], inp2[:, j])
				axs[j, i], caxs[j][i] = plot_spectr(axs[j, i], hist[j][i], cmap=cmap, cscale="log")
				axs[j, i].grid()

		return fig, axs, caxs
